import os
import stat
import base64
import hashlib
import logging
from pathlib import Path


def makeSimpleLogger(name, level=logging.INFO):
    # TODO use extra ...
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()  # FileHander goes to disk
    fmt = ('[%(asctime)s] - %(levelname)8s - '
           '%(name)14s - '
           '%(filename)16s:%(lineno)-4d - '
           '%(message)s')
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = makeSimpleLogger('augpathlib')

LOCAL_DATA_DIR = '.operations'
SPARSE_MARKER = '.sparse'
AUG_XATTR_PREFIX = 'augpathlib'
default_cypher = hashlib.blake2b
cypher_command_lookup = {hashlib.sha256:'sha256sum',
                         hashlib.blake2b:'b2sum'}
red = '\x1b[31m{}\x1b[0m'  # use as red.format(value)


def onerror_windows_readwrite_remove(action, name, exc):
    """ helper for deleting readonly files on windows """
    os.chmod(name, stat.S_IWRITE)
    os.remove(name)


def sysidpath(ignore_options=False, path_class=Path):
    """ get a unique identifier for the machine running this function """
    # in the event we have to make our own
    # this should not be passed in a as a parameter
    # since we need these definitions to be more or less static

    failover = path_class('/var/tmp/machine-id')  # /var/tmp is more persistent than /tmp/

    if hasattr(path_class, 'access'):
        accessf = lambda p: p.access(os.R_OK)
    else:
        # pypy3 3.6 still needs as_poxix here :/
        accessf = lambda p: os.access(p.as_posix(), os.R_OK)

    if not ignore_options:
        options = (
            path_class('/etc/machine-id'),
            failover,  # always read to see if we somehow managed to persist this
        )
        for option in options:
            if (option.exists() and
                accessf(option) and
                option.stat().st_size > 0):
                    return option

    uuid = uuid4()
    with open(failover, 'wt') as f:
        f.write(uuid.hex)

    return failover


def machine_guid(ignore_options=False, path_class=Path):
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                         "SOFTWARE\\Microsoft\\Cryptography", 0,
                         winreg.KEY_READ | winreg.KEY_WOW64_64KEY)
    guid, status = winreg.QueryValueEx(key, 'MachineGuid')
    return guid


if os.name == 'nt':
    import winreg
    def _raw_id(cls, cypher=default_cypher):
        if ((hasattr(cls, '_cache_class') and
             hasattr(cls._cache_class, 'cypher') and
             cls._cache_class.cypher != cypher)):  # FIXME this could be static ...
            cypher = cls._cache_class.cypher

        chunk = machine_guid().encode()
        m = cypher()
        m.update(chunk)
        return m.digest()

else:
    from uuid import uuid4
    def _raw_id(cls):
        return cls(sysidpath(path_class=cls)).checksum()[:16]


def _bind_sysid_(cls):
    if cls.sysid is None:
        cls.sysid = (base64
                     .urlsafe_b64encode(_raw_id(cls))[:-2]
                     .decode())
    else:
        raise ValueError(f'{cls} already has sysid {cls.sysid}')


class FileSize(int):

    @classmethod
    def ofPath(cls, path):
        """ return size of a path alt FileSize.of(path) ? """
        return cls(path.stat().st_size)

    @property
    def mb(self):
        return self / 1024 ** 2

    @property
    def hr(self):
        """ human readable file size """

        def sizeof_fmt(num, suffix=''):
            for unit in ['','K','M','G','T','P','E','Z']:
                if abs(num) < 1024.0:
                    return "%0.0f%s%s" % (num, unit, suffix)
                num /= 1024.0
            return "%.1f%s%s" % (num, 'Yi', suffix)

        if self is not None and self >= 0:
            return sizeof_fmt(self)
        else:
            return '??'  # sigh

    def __str__(self):
        return super().__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__} <{self.hr} {self}>'


class StatResult:
    stat_format = f'\"%n  %i  %o  %s  %w  %W  %x  %X  %y  %Y  %z  %Z  %g  %u  %f\"'

    #stat_format = f'\"\'%n\' %o %s \'%w\' %W \'%x\' %X \'%y\' %Y \'%z\' %Z %g %u %f\"'

    _stat_format_darwin = '\"%N  %i  %k  %z  %SB  %B  %Sa  %a  %Sm  %m  %Sc  %c  %g  %u  %Xp\"'
    # also need -t '%F %T %z' for the %SB and friends, unfortunately it seems nanoseconds is missing?
    # so probably better to check for gstat first  # also -f instead of -c

    def __init__(self, out):
        out = out.decode()
        #name, rest = out.rsplit("'", 1)
        #self.name = name.strip("'")
        #print(out)
        wat = out.split('  ')
        #print(wat)
        #print(len(wat))
        name, ino, hint, size, hb, birth, ha, access, hm, modified, hc, changed, gid, uid, raw_mode = wat

        self.name = name

        def ns(hr):
            date, time, zone = hr.split(' ')
            time, ns = time.split('.')
            return '.' + ns

        self.st_ino = int(ino)
        self.st_blksize = int(hint)
        self.st_size = int(size)
        #self.st_birthtime
        self.st_atime = float(access + ns(ha)) 
        self.st_mtime = float(modified + ns(hm))
        self.st_ctime = float(changed + ns(hc))
        self.st_gid = int(gid)
        self.st_uid = int(uid)
        self.st_mode = int.from_bytes(bytes.fromhex(raw_mode), 'big')

        if hb != '-' and birth != '0':
            self.st_birthtime = float(birth + ns(hb))


class etag:

    cypher = hashlib.md5

    def __init__(self, chunksize):
        self.chunksize = chunksize
        self._m = self.cypher()
        self._parts = []
        self._remainder = b''
        self._last_chunksize = 0

    def update(self, bytes_):
        remainder, self._remainder = self._remainder, b''
        bio = io.BytesIO(remainder + bytes_)  # inefficient for len(bytes_) << self.chunksize
        while True:
            if self._last_chunksize:
                # in the event a restart is required only
                # only read the number of bytes left for
                # the current chunk to update self._m
                chunksize = self.chunksize - self._last_chunksize
                self._last_chunksize = 0
            else:
                chunksize = self.chunksize

            chunk = self._remainder
            self._remainder = bio.read(chunksize)
            if not self._remainder:
                # the second time around at the end this will break the loop if the chunk
                # is exactly equal to the chunk size
                if len(chunk) < chunksize:
                    self._remainder = chunk
                    return

            if chunk:
                self._m.update(chunk)  # empty remainder doesn't advance the function
                digest = self._m.digest()
                self._m = self.cypher()
                if chunksize < self.chunksize:
                    # we are in a restart and need to update the last part with the
                    # digest for the full chunksize
                    self._parts[-1] = digest
                else:
                    self._parts.append(digest)

    def digest(self):
        if self._remainder:
            self._last_chunksize = len(self._remainder)  # in case someone calls an early digest
            self._m.update(self._remainder)
            self._parts.append(self._m.digest())  # do not overwrite self._m yet
            self._remainder = b''  # this allows a stable call at the end

        m = self.cypher()
        m.update(b''.join(self._parts))
        return m.digest(), len(self._parts)

    def hexdigest(self):
        digest, count = self.digest()
        return f'{digest.hex()}-{count}'





class AugPathlibError(Exception):
    """ base class for augpathlib errors """


class UnhandledTypeError(AugPathlibError):
    """ haven't dealt with this yet """


class PathExistsError(AugPathlibError):
    """ path exists so don't try to symlink """


class RemotePathExistsError(AugPathlibError):
    """ the remote path already exists so don't create it """


class DirectoryNotEmptyError(AugPathlibError):  # FIXME compare to PathNotEmpty
    """ directory is not empty """


class TargetPathExistsError(PathExistsError):
    """ when adding to a path if fail_on_exists is set raise this """


class PathNotEmptyError(AugPathlibError):  # FIXME iirc used in clone, compare with DirectoryNotEmpty
    """ folder has children and is not empty, don't overwrite """


class WillNotRemovePathError(AugPathlibError):
    """ we will not even try to remove this path unless you enter the DANGERZONE """


class MetadataIdMismatchError(AugPathlibError):
    """ there is already cached metadata and id does not match """


class MetadataCorruptionError(AugPathlibError):
    """ there is already cached metadata and id does not match """


class NoStreamError(AugPathlibError):
    """ no stream with that name exists similar to file not found or no data """

class NoFileIdError(AugPathlibError):
    """ no file_id """


class NoCachedMetadataError(AugPathlibError):
    """ there is no cached metadata """


class ChecksumError(AugPathlibError):
    """ utoh """


class SizeError(AugPathlibError):
    """ really utoh """


class CommandTooLongError(AugPathlibError):
    """ not the best solution ... """


class NoRemoteMappingError(AugPathlibError):
    """ prevent confusion between local path data and remote path data """


class NotInProjectError(AugPathlibError):
    """fatal: not a spc directory {}"""
    def __init__(self, message=None):
        if message is None:
            more = '(or any of the parent directories)' # TODO filesystem boundaries ?
            self.message = self.__doc__.format(more)


class BootstrappingError(AugPathlibError):
    """ Something went wrong during a bootstrap """


class NoSourcePathError(AugPathlibError):
    """ dictionary at some level is missing the expected key """


class CacheIgnoreError(AugPathlibError):
    """ Path is excluded via cache_ignore. """


class LocalChangesError(AugPathlibError):
    """ File has local changes """


class NoMetadataRetrievedError(AugPathlibError):
    """ we failed to retrieve metadata for a remote id """


class NoRemoteFileWithThatIdError(AugPathlibError):
    """ the file you are trying to reach has been disconnected """


class CacheNotFoundError(AugPathlibError):
    """ no local cache and no remote cache for file

        use this to provide the local path dual of a remote
        for better error reporting """


class WhyDidntThisGetMovedBeforeError(AugPathlibError):
    """ file should already have been moved ... """


class RepoExistsError(AugPathlibError):
    """ a repository already exists at this path """


class NotInRepoError(AugPathlibError):
    """ a repository does not exist in this
        or any of the parents of this path """


class NoCommitsForFile(AugPathlibError):
    """ could not find any commits for file """


class InvalidRefError(AugPathlibError):
    """ ref.is_valid() returned False """


class RemoteAlreadyAnchoredError(AugPathlibError):
    """ a Remote class has already been anchored locally """


class CacheExistsError(AugPathlibError):
    """ attempting to init a cache that already exists """


class FileHasNotChangedError(AugPathlibError):
    """ signal cases where a file has not changed
        and thus that no action will be taken """



import os
import pathlib
import unittest
import pytest
import augpathlib as aug
from augpathlib import RepoPath, LocalPath, exceptions as exc
from .common import onerror, skipif_no_net, temp_path

testing_base = RepoPath(temp_path)


class HybridPath(RepoPath, LocalPath):
    """ Combined functionality """


HybridPath._bind_flavours()


class TestRepoPath(unittest.TestCase):
    def setUp(self):
        if not testing_base.exists():
            testing_base.mkdir()

    def tearDown(self):
        RepoPath(testing_base).rmtree(onerror=onerror)

    def test_init(self):
        rp = testing_base / 'test-repo'
        repo = rp.init()
        assert repo, f'hrm {rp!r} {repo}'
        return rp

    def test_clone_path(self):
        rp = testing_base.clone_path('git@github.com:tgbugs/augpathlib.git')
        assert rp.name == 'augpathlib', 'ssh form failed'

        rp = testing_base.clone_path('https://github.com/tgbugs/augpathlib.git')
        assert rp.name == 'augpathlib', 'https form failed'

        rp = testing_base.clone_path('~/git/augpathlib')
        assert rp.name == 'augpathlib', 'local form failed'

    def test_init_from_local_repo(self):
        rp = testing_base / 'test-repo'
        this_repo_path = RepoPath(__file__).working_dir
        if rp.working_dir is not None:
            pytest.skip('not testing inside another git repo')

        if this_repo_path is None:
            pytest.skip('this test file is not under version control, so there is no local repo')

        else:
            this_repo = this_repo_path.repo
            repo = rp.init(this_repo_path, depth=1)
            assert repo, f'{rp!r} {repo}'
            return rp

    @skipif_no_net
    def test_init_with_remote(self):
        rp = testing_base / 'test-repo'
        repo = rp.init('https://github.com/tgbugs/augpathlib.git', depth=1)
        assert repo, f'{rp!r} {repo}'
        return rp

    @skipif_no_net
    def test_clone_from(self):
        rp = testing_base.clone_from('https://github.com/tgbugs/augpathlib.git', depth=1)
        assert rp.repo, f'{rp!r} {rp.repo}'
        return rp

    def test_show(self):
        test_byte1 = b'\x98'
        test_byte2 = b'\x99'
        tag1 = 'tag1'
        tag2 = 'tag2'
        rp = testing_base / 'test-repo-show'
        rp.init()
        rp.repo.git.commit(message='Inital commit', allow_empty=True)

        tf = rp / 'test-file'
        with open(tf, 'wb') as f:
            f.write(test_byte1)

        tf.commit_from_working_tree("test byte one")
        tf.repo.create_tag(tag1)

        with open(tf, 'wb') as f:
            f.write(test_byte2)

        tf.commit_from_working_tree("test byte two")
        tf.repo.create_tag(tag2)

        value = tf.show('tag1')
        assert value == test_byte1

    def test_stale_repo_cache(self):
        rp = self.test_init()
        rp.repo
        aug.AugmentedPath(rp).rmtree()  # have to call rmtree that won't invoke repo.close()
        try:
            rp.repo
            assert False, 'should have failed'
        except exc.NotInRepoError:
            pass

        rp2 = RepoPath(rp)
        try:
            rp2.repo
            assert False, 'should have failed'
        except exc.NotInRepoError:
            pass

    @skipif_no_net
    def test_remote_uris(self):
        rp = self.test_init_with_remote()
        h = rp.remote_uri_human()
        m = rp.remote_uri_machine()
        h = rp.remote_uri_human('master')
        m = rp.remote_uri_machine('master')
        assert h != m  # TODO probably need a better test here

    def test_repo_context(self):
        rp = testing_base / 'test-repo-context'
        rp.init()
        rp.repo.git.commit(message='Inital commit', allow_empty=True)
        ref_target = 'ref-target'
        rp.repo.create_head(ref_target)

        test_byte1 = b'\x98'
        test_byte2 = b'\x99'
        test_byte3 = b'\x97'
        test_byte4 = b'\x96'

        tc = rp / 'test-committed'
        with open(tc, 'wb') as f:
            f.write(test_byte1)

        tc.commit_from_working_tree("test byte one")

        tu = rp / 'test-uncommitted'
        with open(tu, 'wb') as f:
            f.write(test_byte2)

        with rp.repo.getRef(ref_target):
            # this will cause an error I'm sure ...
            # will have to reorder the stack in this case
            # numerous types of errors we can expect here
            '''
            git.exc.GitCommandError: Cmd('git') failed due to: exit code(1)
              cmdline: git checkout master
              stderr: 'error: The following untracked working tree files would be overwritten by checkout:
            	test-committed
            Please move or remove them before you switch branches.
            Aborting'
            '''

            with open(tc, 'wb') as f:
                f.write(test_byte3)

            with open(tu, 'wb') as f:
                f.write(test_byte4)


class TestComplex(unittest.TestCase):
    test_file = 'test-file'

    def setUp(self):
        self.hp = HybridPath(testing_base).clone_path('test-repo')
        self.hp.init()
        self.test_file = self.hp / self.test_file

    def tearDown(self):
        RepoPath(testing_base).rmtree(onerror=onerror)

    def test_commit(self):
        self.test_file.touch()
        self.test_file.add_index()
        c1 = self.test_file.commit(message='test commit 1')
        self.test_file.data = (b'a' for _ in (0,))
        self.test_file.add_index()
        c2 = self.test_file.commit(message='test commit 1')
        self.test_file.data = (b'b' for _ in (0,))

    def test_diff(self):
        self.test_commit()
        d = self.test_file.diff('HEAD', 'HEAD~1')
        assert d, f'd'

    def test_working_dir(self):
        [RepoPath._repos.pop(k) for k in list(RepoPath._repos)]
        rp = RepoPath(str(self.test_file))
        assert rp.working_dir is not None, f'wat {rp}'
        assert rp.repo is not None, f'wat {rp} {rp.working_dir}'



import os
import pathlib
import unittest
import pytest
import augpathlib as aug
from augpathlib import RepoPath, LocalPath, exceptions as exc
from .common import onerror, skipif_no_net, temp_path, this_file


def test():
    _zp =  this_file.parent / 'test.zip'
    # TODO create the zip we will test using mkdir, touch, and data = etc.
    zp = aug.ZipPath(_zp)

    zrp = zp.path_relative_zip
    # FIXME the internal paths should
    # be hidden as an implementation detail

    rc = sorted(zrp.rchildren)
    assert all([c.exists() for c in rc])
    grid = [(c.exists(), c.is_dir(), c)
            for c in sorted(zp.rchildren)]
    c = rc[-1]

    # TODO open_stream ...
    # to get actual generic behavior on open

    with c.open() as f:
        test1 = f.read()

    czp = zp / c
    with czp.open() as f:
        test2 = f.read()

    assert test1 == test2


if __name__ == '__main__':
    test()



import pathlib
import warnings
from augpathlib import exceptions as exc
from augpathlib.meta import PathMeta
from augpathlib.core import AugmentedPath, EatPath
from augpathlib.utils import log, default_cypher
from augpathlib.utils import LOCAL_DATA_DIR, SPARSE_MARKER
from augpathlib import remotes


class CachePath(AugmentedPath):
    # CachePaths this needs to be a real path so that it can navigate the local path sturcture
    # FIXME Not sure I believe that, given the tradeoff
    """ Local data about remote objects.
        This is where the mapping between the local id (aka path)
        and the remote id lives. In a git-like world this is the
        cache/index/whatever we call it these days

        This is the bridge class that holds the mappings.
        Always start bootstrapping from one of these classes
        since it has both the local and remote identifiers,
        and therefore can be called and used before specifying
        the exact implementations for the local and remote objects.
    """

    _local_data_dir = LOCAL_DATA_DIR
    _sparse_marker = SPARSE_MARKER
    cache_ignore = _local_data_dir, _sparse_marker, '.git', # TODO

    _local_class = None
    _remote_class_factory = None

    _backup_cache = None
    _not_exists_cache = None

    def __enter__(self):
        if self.is_dir():
            self._entered_from = self.local.cwd()  # caches can't exist outside their anchor anymore
            self.chdir()
            return self
        else:
            super().__enter__(self)

    @classmethod
    def setup(cls, local_class, remote_class_factory):
        """ call this once to bind everything together """

        cn = self.__class__.__name__
        warnings.warn(f'{cn}.setup is deprecated please switch to RemotePath._new',
                      DeprecationWarning,
                      stacklevel=2)

        cls._local_class = local_class
        cls._remote_class_factory = remote_class_factory
        local_class._cache_class = cls
        remote_class_factory._cache_class = cls

        # a nice side effect of weighing anchor here is that it
        # enforces order of operations for setup then init etc.
        if hasattr(cls, '_anchor'):
            cls.weighAnchor()

    @classmethod
    def weighAnchor(cls):
        # return a value to indicate that there was an anchor since
        # we no longer error when already underway
        acls = cls._abstract_class()
        if hasattr(acls, '_anchor'):
            return delattr(acls, '_anchor')

    def anchorClassHere(self, remote_init=True):
        """ Use this to initialize the class level anchor from an instance. """

        # FIXME WARNING you can shoot yourself in the foot with this if
        # there is another anchor further up the tree from this one
        # FIXME further, this means that there has to have been a way
        # to construct a live CachePath by sideloading the remote id
        # which is OK for the path where local/remote binding has already
        # been completed
        if not hasattr(self.__class__, '_anchor'):
            self.__class__._abstract_class()._anchor = self
            self.local_data_dir_init()  # call every time for insurance
            self._remote_class.anchorToCache(self, init=remote_init)

        else:
            raise ValueError(f'{self.__class__} already anchored to {self.__class__._anchor}')

    @property
    def local_class(self):
        if self.is_helper_cache:
            return self._cache_parent.local_class

        return self._local_class

    def __new__(cls, *args, meta=None, remote=None, **kwargs):
        # TODO do we need a version of this for the locals
        # and remotes? I don't think we create 'alternate' remotes or locals ...

        self = super().__new__(cls, *args, **kwargs)

        # clone any existing locals and remotes
        if args:
            path = args[0]
            if isinstance(path, CachePath):
                self._cache_parent = path
                if hasattr(self._cache_parent, '_in_bootstrap'):
                    # it is ok to do this and not clean up because
                    # child caches are ephemoral
                    self._in_bootstrap = self._cache_parent._in_bootstrap

                if path.local is not None:  # this might be the very fist time local is called
                    pass  # we don't use self._local anymore too many issues

            elif isinstance(path, remotes.RemotePath):
                #self._remote = path
                #self.meta = path.meta
                # in order for this to work the remote has to already
                # know where the cache should live, which it doesn't
                # use move instead for cases where the semantics are well defined
                raise TypeError('Not entirely sure what to do in this case ...')

        return self

    def __init__(self, *args, meta=None, remote=None, **kwargs):
        if remote:
            self._remote = remote
            self._meta_setter(remote.meta)
        elif meta:
            self._meta_updater(meta)
        else:
            path = args[0]
            if self.meta is None:
                raise exc.NoCachedMetadataError(self.local)

            elif isinstance(path, LocalPath):
                path._cache = self

        super().__init__()

    @property
    def anchor(self):
        raise NotImplementedError('You need to define the rule for determining '
                                  'the local cache root for \'remote\' paths. '
                                  'These are sort of like pseudo mount points.')

    @property
    def trash(self):
        raise NotImplementedError('implement in subclasses')
        # FIXME mkdir and put it in a more conventional location

    @property
    def _trashed_path(self):
        return self.trash / f'{self.parent.id}-{self.id}-{self.name}'

    @property
    def _trashed_path_short(self):
        return self.trash / self.name  # FIXME SIGH

    def crumple(self):  # FIXME change name to something more obvious ...
        try:
            self.rename(self._trashed_path)
        except OSError as e:
            if e.errno == 36:  # File name too long  # SIGH
                log.critical(f'Had to rename trash {self._trashed_path} -> {self._trashed_path_short}')
                self.rename(self._trashed_path_short)
            else:
                raise e

        return trashed

    @property
    def local_data_dir(self):
        return self.anchor.local / self._local_data_dir

    @property
    def local_objects_dir(self):
        """ sort of like .git/objects """
        return self.local_data_dir / 'objects'

    @property
    def local_object_cache_path(self):
        # FIXME probably need the 2 char directory convention
        # to limit directory size
        return self.local_objects_dir / self.cache_key

    def local_data_dir_init(self, exist_ok=True, symlink_objects_to=None):
        # FIXME shouldn't this always run once the
        # first time a class is initialized and then
        # modify new to never call it again?
        self.local_data_dir.mkdir(exist_ok=exist_ok)

        lod = self.local_objects_dir
        if symlink_objects_to is not None:
            # NOTE this interacts with anchorClassHere
            # we do want to always run this local_data_dir_init
            # to avoid hard to debug errors, which means that if
            # a user wants to override, we have to rmdir first
            # NOTE we do NOT rmtree here, the user needs to to that
            # explicitly so we don't accidentally remove everything
            # NOTE this whole design is dumb, and in cases where
            # there are just remote objects we should be stashing
            # them in ~/.cache or similar
            if self.local_objects_dir.is_symlink():
                msg = f'{lod} is already symlinked to {lod.readlink()}'
                raise NotADirectoryError(msg)
            elif lod.exists():
                lod.rmdir()

            lod.symlink_to(symlink_objects_to)
        else:
            lod.mkdir(exist_ok=exist_ok)

        try:
            self.trash.mkdir(exist_ok=exist_ok)
        except NotImplementedError:
            # if there's no trash, there's no trash
            pass

    @property
    def is_helper_cache(self):
        return hasattr(self, '_cache_parent')

    def __truediv__(self, key, update_meta=True):
        # basically RemotePaths are like relative CachePaths ... HRM
        # they are just a name and an id ... the id of their parent
        # root needs to match the id of the cache ... which it usually
        # does by construction
        parent = self.parent if self.parent.meta is not None else self  # FIXME should be in def parent ???
        if isinstance(key, remotes.RemotePath):
            # FIXME not just names but relative paths???
            remote = key
            try:
                child = self._make_child(
                    remote._parts_relative_to(self.remote, parent),
                    remote, update_meta=update_meta)
            except AttributeError as e:
                raise exc.AugPathlibError('aaaaaaaaaaaaaaaaaaaaaa') from e

            return child

        elif isinstance(key, str):
            child = self.local / key
            if child.exists() or child.is_broken_symlink():
                return child.cache
            else:
                raise FileNotFoundError('There is no local cached file with that name. Cannot construct cache.')

        else:
            pass  # error incoming

        raise TypeError('Cannot construct a new CacheClass from an object '
                        f'without an id and a name! {key}')

    def __rtruediv__(self, cache):
        """ key is a subclass of self.__class__ """
        # I assume that this happens when a cache is constructed from
        # an relative cache?
        out = self._from_parts([cache.name] + self._parts, init=False)
        out._init()
        cache.remote._cache_setter(out)  # this seems more correct?
        #out._meta_setter(cache.meta)
        return out

    def _make_child(self, args, remote, update_meta=True):
        drv, root, parts = self._parse_args(args)
        drv, root, parts = self._flavour.join_parsed_parts(
            self._drv, self._root, self._parts, drv, root, parts)
        child = self._from_parsed_parts(drv, root, parts, init=False)  # short circuits
        child._init()
        if isinstance(remote, remotes.RemotePath):
            remote._cache_setter(child, update_meta=update_meta)
        else:
            raise ValueError('should not happen')

        return child

    def bootstrap(self, meta, *,
                  parents=False,
                  recursive=False,
                  fetch_data=False,
                  size_limit_mb=2,
                  only=tuple(),
                  skip=tuple(),
                  sparse=tuple(),):
        try:
            self._in_bootstrap = True
            return  list(self._bootstrap(meta,
                                         parents=parents,
                                         recursive=recursive,
                                         fetch_data=fetch_data,
                                         size_limit_mb=size_limit_mb,
                                         only=only,
                                         skip=skip,
                                         sparse=sparse,))
        finally:
            delattr(self, '_in_bootstrap')
            if hasattr(self, '_meta'):
                delattr(self, '_meta')

            if hasattr(self, '_is_sparse_root'):
                self._mark_sparse()
                delattr(self, '_is_sparse_root')

    def _sparse_root(self):  # TODO consider SparseHelper
        return self != self.parent and (self / self._sparse_marker).exists()

    def is_sparse(self):
        return self._sparse_root() is not None

    def _clear_sparse(self):
        mark = self.local / self._sparse_marker
        mark.unlink()

    def _mark_sparse(self):
        """ default implementation for marking folders as sparse

            this uses a file in the folder, but a better implementation
            if one has access to xattrs is to use those instead

            as such an implementation of _mark_sparse is also provided
            on EatPath """
        mark = self.local / self._sparse_marker
        mark.touch()

    def _meta_is_root(self, meta):
        """ used to identify the root during bootstrap """
        raise NotImplementedError('implement in subclass')

    def _bootstrap(self, meta, *,
                   parents=False,
                   fetch_data=False,
                   size_limit_mb=2,
                   recursive=False,
                   only=tuple(),
                   skip=tuple(),
                   sparse=tuple(),):
        """ The actual bootstrap implementation """

        # figure out if we are actually bootstrapping this class or skipping it
        if not meta or meta.id is None:
            raise exc.BootstrappingError(f'PathMeta to bootstrap from has no id! {meta}')

        if only or skip or sparse:
            if self._meta_is_root(meta):
                # since we only go one organization at a time right now
                # we never want to skip the top level id
                log.info(f'Bootstrapping {meta.id} -> {self.local!r}')
            elif meta.id in skip:
                log.info(f'Skipped       {meta.id} since it is in skip')
                return
            elif only and meta.id not in only:
                log.info(f'Skipped       {meta.id} since it is not in only')
                return
            else:
                if sparse and meta.id in sparse:
                    log.info(f'Sparse strap {meta.id} -> {self.local!r}')
                    self._is_sparse_root = True
                    sparse = True
                else:
                    # if you pass the only mask so do your children
                    log.info(f'Bootstrapping {meta.id} -> {self.local!r}')

                only = tuple()

        if self.meta is not None and not recursive:
            msg = f'{self} already has meta!\n{self.meta.as_pretty()}'
            raise exc.BootstrappingError(msg)

        if self.exists() and self.meta and self.meta.id == meta.id:
            self._meta_updater(meta)

        else:
            # set single use bootstrapping id
            self._bootstrapping_id = meta.id

            # directory, file, or fake file as symlink?
            is_file_and_fetch_data = self._bootstrap_prepare_filesystem(parents,
                                                                        fetch_data,
                                                                        size_limit_mb,
                                                                        sparse,)

            is_file_and_fetch_data = False  # XXX NOTE _bootstrap_prepare_filesystem always returns None
            # remove this static assignment to False if there is a use case for bootstrapping the data
            self._bootstrap_data(is_file_and_fetch_data)

        if recursive:  # ah the irony of using loops to do this
            yield from self._bootstrap_recursive(only, skip, sparse)

        yield self

    def _bootstrap_recursive(self, only=tuple(), skip=tuple(), sparse=False):
        # TODO if rchildren looks like it could be bad
        # go back up to dataset level?
        #sname = lambda gen: sorted(gen, key=lambda c: c.name)  # c.name doesn't work for remotes
        #rcs = sname(self.remote._rchildren(create_cache=False, sparse=sparse))
        rcs = self.remote._rchildren(create_cache=False, sparse=sparse)

        local_paths = list(self.local.rchildren)
        local_files = set(p for p in local_paths if p.is_file() or p.is_broken_symlink())
        file_index = {f.cache_id:f for f in local_files}  # FIXME WARNING can get big
        # FIXME have to compute file_index here because for some reason
        # computing local_dirs will remove folders entirely !??
        local_dirs = set(p.relative_to(self.anchor) for p in local_paths if p.is_dir())
        if local_dirs:
            rcs = list(rcs)  # sigh
            remote_dirs = set(c for c in rcs if c.is_dir())

            rd = set(d.as_path() for d in remote_dirs)  # FIXME as_path => lots of network calls
            old_local = local_dirs - rd
            while old_local:
                thisl = sorted(old_local, key=lambda d: len(d.as_posix()))
                for d in thisl:
                    ad = self.anchor.local / d
                    if ad.cache is None:
                        log.critical(f'would you fix the nullability already? {d}')
                        continue
                    new = ad.cache.refresh()
                    #log.info(f'{new}')
                    local_dirs = set(ld for ld in local_dirs
                                     if not ld.as_posix().startswith(d.as_posix()))
                    old_local = local_dirs - rd

        if sparse:
            #if local_dirs:
                #gen = (c for c in _local_remotes if c.is_dir() or (c.is_file() and c._sparse_include()))
            #else:
            gen = (c for c in rcs if c.is_dir() or (c.is_file() and c._sparse_include()))
            # FIXME rcs still takes too long, though using the generator
            # does get some useful work done first
        else:
            # FIXME horrid performance on remotes with loads of files
            gen = sorted(rcs, key=lambda c: len(c.as_path().as_posix()))

        for child in gen:
            # use the remote's recursive implementation
            # not the local implementation, since the
            # remote may have additional requirements
            #child.bootstrap(only=only, skip=skip)
            # because of how remote works now we don't even have to
            # bootstrap this
            cc = child.cache

            if cc is None:
                if child.is_file() and child.id in file_index:
                    _cache = file_index[child.id].cache
                    cmeta = _cache.meta
                    rmeta = child.meta
                    file_is_different, nmeta = self._update_meta(cmeta, rmeta)
                    if file_is_different:
                        log.critical(f'WAT {_cache}')
                    else:
                        yield _cache
                        # yield the old cache if it exists
                        # otherwise consumers of bootstrap will
                        # think the file may have been deleted
                        continue

                cc = child.cache_init()
                log.debug(cc)

            yield cc

    def _bootstrap_prepare_filesystem(self, parents, fetch_data, size_limit_mb, sparse=False):
        # we could use bootstrapping id here and introspect the id, but that is cheating
        if self.remote.is_dir():
            if not sparse and not self.exists():
                # the bug where this if statement put in as an and is a really
                # good example of how case/cond etc help you reasona about what
                # a block of branches is really doing -- this one was implementing
                # a covering set which is not obvious if implemented this way
                # you could do this with a dict or something else in pythong
                # bit it is awkward (see also my crazy case implementation in interlex)
                self.mkdir(parents=parents)

        elif self.remote.is_file():
            if sparse and not self._sparse_include():
                return

            if not self.parent.exists():
                self.parent.mkdir(parents=parents)

            toucha_da_filey = (fetch_data and
                               self.meta.size is not None and
                               self.meta.size.mb < size_limit_mb)

            if toucha_da_filey:
                self.touch()
                # running this first means that we will use xattrs instead of symlinks
                # this is a bit opaque, but since meta uses a setter we can't pass a
                # param to make it clear (oh look, python being dumb again!)
            else:
                pass  # we are using symlinks

        else:
            raise BaseException(f'Remote is not a file or directory {self}')

    def _bootstrap_data(self, is_file_and_fetch_data=False):
        """ XXX UNUSED """
        if is_file_and_fetch_data:
            if self.remote.meta.size is None:
                self.remote.refresh(update_cache=True)

            self.local.data = self.remote.data
            # with open -> write should not cause the inode to change

            self.validate_file()

    def _sparse_include(self):
        raise NotImplementedError('implement in subclass')

    def validate_file(self):
        meta = self.meta
        if meta.etag:
            local_checksum, local_count = self.local.etag(meta.chunksize)
            cache_checksum, cache_count = meta.etag
            if local_checksum != cache_checksum or local_count != cache_count:
                msg = (f'etags do not match!\n(!='
                       f'\n{local_checksum}-{local_count}'
                       f'\n{cache_checksum}-{cache_count}\n)')
                log.critical(msg)

        elif meta.checksum:
            lc = self.local.meta.checksum
            cc = self.meta.checksum
            if lc != cc:
                msg = f'Checksums do not match!\n(!=\n{lc}\n{cc}\n)'
                log.critical(msg)  # haven't figured out how to comput the bf checksums yet
                #raise exc.ChecksumError(msg)
        elif meta.size is not None:
            log.warning(f'No checksum! Your data is at risk!\n'
                        f'{self.remote!r} -> {self.local!r}! ')
            ls = self.local.meta.size
            cs = self.meta.size
            if ls != cs:
                raise exc.SizeError(f'Sizes do not match!\n(!=\n{ls}\n{cs}\n)')
        else:
            log.warning(f'No checksum and no size! Your data is at risk!\n'
                        '{self.remote!r} -> {self.local!r}! ')

    @property
    def remote(self):
        if hasattr(self, '_remote'):
            return self._remote

        if hasattr(self, '_cache_parent'):
            return self._cache_parent.remote

        id = self.id  # bootstrapping id is a one time use so keep it safe
        if id is None:  # zero is a legitimate identifier
            return

        anchor = self.anchor
        if anchor is None:  # the very first ...
            # in which case we need the id for factory AND class
            self._bootstrapping_id = id  # so we set it again
            anchor = self  # could double check if the id has the info too ...

        # FIXME remove?
        if self._remote_class_factory is not None or (hasattr(self, '_remote_class') and
                                                      self._remote_class is not None):
            # we don't have to have a remote configured to check the cache
            if not hasattr(self, '_remote_class'):
                #log.debug('rc')
                # NOTE there are many possible ways to set the anchor
                # we need to pick _one_ of them
                self._remote_class = self._remote_class_factory(anchor,
                                                                self.local_class)

            if (hasattr(self._remote_class, '_api_class') and
                not hasattr(self._remote_class, '_api')):
                # FIXME SIGH this should be one shot not a hasattr check
                # every first time we have to get the remote for a cache
                self._remote_class.anchorToCache(anchor)

            if not hasattr(self, '_remote'):
                self._remote = self._remote_class(id, cache=self)

            return self._remote

    @property
    def local(self):
        local = self.local_class(self)
        if self.is_helper_cache:
            cache = self._cache_parent
        else:
            cache = self

        local._cache = cache
        return local

    def dedupe(self, other, pretend=False):
        # FIXME blackfynn doesn't set update when a folder name changes ??!
        if self.id != other.id:
            raise ValueError('Can only dedupe when ids match, {self.id} != {other.id}')

        su, ou = self.meta.updated, other.meta.updated
        lsu, lou = self.local.meta.updated, other.local.meta.updated
        if su < ou:
            old, new = self, other

        elif su > ou:
            new, old = self, other

        elif lsu is None and lou is None:
            new, old = self, other

        elif lsu is None:
            old, new = self, other

        elif lou is None:
            new, old = self, other

        elif lsu < lou:
            old, new = self, other

        elif lsu > lou:
            new, old = self, other

        else:  # ==
            ss, os = self.meta.size, other.meta.size
            if ss is not None and os is not None:
                new, old = self, other

            elif ss is None:
                old, new = self, other

            elif os is None:
                new, old = self, other

            else:
                raise BaseException('how did we get here!?')

        file_is_different, meta = self._update_meta(old.meta, new.meta)
        if file_is_different:
            log.info(f'{self}\n!=\n{other}\n{meta}')

        if not pretend:
            #old.rename('/dev/null')  # hah
            pass

        return new
        # TODO go look in meta for this
        # check updated ... etc.
        # missing size
        # missing file_id

    @property
    def id(self):
        if not hasattr(self, '_id'):  # calls to self.exists() are too expensive for this nonsense
            if self.meta:
                self._id = self.meta.id
                return self._id

            elif hasattr(self, '_bootstrapping_id'):
                id = self._bootstrapping_id
                delattr(self, '_bootstrapping_id')  # single use only
                return id
            else:
                return

        return self._id

    @property
    def cache_key(self):
        """ since some systems have compound ids ... """
        raise NotImplementedError

    # TODO how to toggle fetch from remote to heal?
    @property
    def meta(self):
        raise NotImplementedError

        if hasattr(self, '_meta'):
            return self._meta  # for bootstrap

    def _meta_setter(self, pathmeta, memory_only=False):
        """ so much for the pythonic way when the language won't even let you """
        if not memory_only:
            raise TypeError('You must explicitly set memory_only=True to use this '
                            'otherwise you risk dataloss.')

        if self.meta and self.id != pathmeta.id:
            raise exc.MetadataIdMismatchError('Cache id does not match meta id! '
                                              f'{self.id} != {pathmeta.id}\n{pathmeta}')

        self._meta = pathmeta

    def recover_meta(self):
        """ rebuild restore reconnect """

        root = self.parent.local.find_cache_root()
        if root is None:
            #breakpoint()
            raise exc.NotInProjectError(f'{self.parent.local} is not in a project!')

        breakpoint()
        raise NotImplementedError()
        children = list(self.parent.remote.children)  # if this is run from dismatch meta we have issues
        isf = self.is_file()
        isd = self.is_dir()
        candidates = []
        def inner(child):
            if child.is_dir() and isd:
                if child.name == self.name:
                    self.meta = child.meta
                    return

            elif child.is_file() and isf:
                log.debug(f'{child.name} {child.stem}, {child.suffix!r}')
                log.debug(f'{self.name} {self.stem}, {self.suffix!r}')
                if child.name == self.name:
                    self.meta = child.meta
                elif child.name == self.stem:
                    candidates.append(child)
                elif child.stem == self.name:
                    candidates.append(child)
                elif child.stem == self.stem:
                    # worst cases
                    candidates.append(child)

            else:
                #log.critical('file type mismatch')
                pass

        for child in children:
            inner(child)
            # it looks like if we do fail over to retrieving a package it does go to files
            # so this is an ok approach and we don't have to deal with that at this level
        if not candidates:
            wat = '\n'.join(c.name for c in children)
            message = (f'We seem to have lost {self.parent} -/-> {self.name}'
                       f'\n{self.parent.uri_human}\n{wat}\n{self.name}')
            log.critical(message)
            dataset = self.dataset
            maybe = []
            for c in self.dataset.remote.rchildren:
                if c.parent and c.parent.id == self.parent.id or c.stem == self.stem:
                    maybe.append(c)

            [inner(m) for m in maybe]
            #candidates
            #dataset.bootstrap(dataset.meta, recursive=True)
            #raise exc.NoRemoteMappingError

        elif len(candidates) == 1:
            remote = candidates[0]
            log.critical('How did we write this file without moving it beforhand?!?\n'
                         f'{self.local} -/-> {remote.name}')
            self.meta = remote.meta  # go ahead and set this even though we change?
            self.move(remote=remote)
        else:
            raise BaseException('multiple candidates!')

    def refresh(self, update_data=False, size_limit_mb=2, force=False):
        if self.meta is None:
            breakpoint()

        limit = (size_limit_mb if
                 not self.meta.size or (size_limit_mb > self.meta.size.mb)
                 else self.meta.size.mb + 1)
        new = self.remote.refresh(update_cache=True,
                                  update_data=update_data,
                                  update_data_on_cache=(self.is_file() and self.exists()),
                                  size_limit_mb=size_limit_mb,
                                  force=force)
        if new is not None:
            return new
        else:
            log.info(f'Remote for {self} has been deleted. Moving to trash.')
            try:
                self.rename(self.trash / f'{self.parent.id}-{self.id}-{self.name}')
            except FileNotFoundError as e:
                if not self.trash.exists():
                    self.trash.mkdir()
                    log.info('created {self.trash}')
                else:
                    raise e

    def fetch(self, size_limit_mb=2):
        """ bypass remote to fetch directly based on stored meta """
        meta = self.meta
        if self.is_dir():
            raise NotImplementedError('not going to fetch all data in a dir at the moment')
        if meta.file_id is None:
            self.refresh(update_data=True, force=True)
            # the file name could be different so we have to return here
            return

        size_ok = size_limit_mb is not None and meta.size is not None and meta.size.mb < size_limit_mb
        size_not_ok = size_limit_mb is not None and meta.size is not None and meta.size.mb > size_limit_mb

        if size_ok or size_limit_mb is None:  # FIXME should we force fetch here by default if the file exists?
            if self.is_broken_symlink():
                # FIXME touch a temporary file and set the meta first!
                self.unlink()
                self.touch()
                self._meta_setter(meta)

            log.info(f'Fetching remote via cache id {self.id} -> {self.local}')
            self.local.data = self.data  # note that this should trigger storage to .ops/objects

        if size_not_ok:
            log.warning(f'File is over the size limit {meta.size.mb} > {size_limit_mb}')

    def move(self, *, remote=None, target=None, meta=None):
        """ instantiate a new cache and cleanup self because we are moving """
        # FIXME what to do if we have data
        if remote is None and (target is None or meta is None):
            raise TypeError('either remote or meta and target are required arguments')

        # deal with moving to a different directory that might not even exist yet
        if target is None:
            if not isinstance(self.anchor, self.__class__):
                raise TypeError(f'mismatched anchor types {self!r} {self.anchor!r}')

            target = self.anchor / remote  # FIXME why does this not try to instantiate the caches? or does it?

        if target.absolute() == self.absolute():
            log.warning(f'trying to move a file onto itself {self.absolute()}')
            return target

        common = self.commonpath(target).absolute()
        target_parent = target.parent.absolute()
        parent = self.parent.absolute()

        assert target.name != self.name or target_parent != parent

        if target_parent != parent:
            _id = remote.id if remote else meta.id
            log.warning('A parent of current file has changed location!\n'
                        f'{common}\n{self.relative_to(common)}\n'
                        f'{target.relative_to(common)}\n{_id}')


        if not target_parent.exists():
            if remote is None:  # we have to have a remote to pull parent structure
                remote = self._remote_class(meta)

            target_parent.mkdir_cache(remote)

        do_cast = not isinstance(target, self.__class__)
        if do_cast:
            target = self.__class__(target, meta=meta)

        if target.exists() or target.is_broken_symlink():
            if target.id == self.id: #(remote.id if remote else meta.id):
                if self.is_broken_symlink():
                    # we may be a package with extra metadata that needs to
                    # be merged with the target before we are unlinked
                    file_is_different = target._meta_updater(self.meta)
                    # FIXME ... if file is different then this causes staleness
                    # and we need to fetch
                    if file_is_different:
                        log.critical('DO SOMETHING ABOUT THIS STALE DATA'
                                     f'\n{target}\n{target.meta.as_pretty()}')

                elif do_cast:
                    # the target meta was just put there, if the ids match it should be ok
                    # however since arbitrary meta can be passed in, best to double check
                    file_is_different = target._meta_updater(self.meta)
                    if file_is_different:
                        log.critical('Something has gone wrong'
                                     f'\n{target}\n{target.meta.as_pretty()}')
                else:
                    # directory moves that are resolved during pull
                    log.warning(f'what is this!?\n{target}\n{self}')
            elif target.is_broken_symlink():
                remote._cache = self  # restore the mapping for remote -> self
                raise exc.WhyDidntThisGetMovedBeforeError(f'\n{target}\n{self}')
            else:
                raise exc.PathExistsError(f'Target {target} already exists!')

        if self.exists():
            safe_unlink = target.local.parent / f'.unlink-{target.name}'
            try:
                if target.is_broken_symlink():
                    target.rename(safe_unlink)

                self.rename(target)  # if target is_dir then this will fail, which is ok
            except BaseException as e:
                log.exception(e)
                if safe_unlink.is_broken_symlink():
                    safe_unlink.rename(target)
            finally:
                if safe_unlink.is_broken_symlink():
                    safe_unlink.unlink()

        elif self.is_broken_symlink():
            # we don't move to trash here because this was just a file rename
            self.unlink()  # don't move the meta since it will break the naming insurance measure

        return target

    def __repr__(self):
        local = repr(self.local) if self.local else 'No local??' + str(self)
        remote = (f'{self.remote.__class__.__name__}({self.id!r})'
                  if self.remote else str(self.id))
        return self.__class__.__name__ + ' <' + local + ' -> ' + remote + '>'

    @property
    def data(self):
        raise NotImplementedError('implement in subclass')


CachePath._bind_flavours()


class ReflectiveCache(CachePath):
    """ Oh, it's me. """

    @property
    def meta(self):
        return self.local.meta


ReflectiveCache._bind_flavours()


class EatCache(EatPath, CachePath):
    xattr_prefix = None

    @property
    def meta(self):
        if self.exists():
            xattrs = self.xattrs()
            pathmeta = PathMeta.from_xattrs(xattrs, self.xattr_prefix, self)
            return pathmeta

    def _meta_setter(self, pathmeta, memory_only=False):
        #log.warning(f'!!!!!!!!!!!!!!!!!!!!!!!!2 {self}')
        # TODO cooperatively setting multiple different cache types?
        # do we need to use super() or something?

        if self.exists():
            if self.is_symlink():
                raise TypeError('will not write meta on symlinks! {self}')
            # FIXME FIXME FIXME this needs to be written to absolutely
            # prevent the writing of new metadata onto an existing file
            # where the checksum differs, the old version needs to be
            # trashed before any of this is written, otherwise the old
            # metadata is lost >_<
            self.setxattrs(pathmeta.as_xattrs(self.xattr_prefix))
            if hasattr(self, '_meta'):  # prevent reading from in-memory store
                delattr(self, '_meta')

        else:
            # the glories of the inconsistencies and irreglarities of python
            # you can't setattr using super() so yes you _do_ actually have to
            # implement a setter sometimes >_<
            super()._meta_setter(pathmeta, memory_only=memory_only)


EatCache._bind_flavours()


class SqliteCache(CachePath):
    """ a persistent store to back up the xattrs if they get wiped """

    def __init__(self, *args, meta=None, **kwargs):
        if meta is not None:
            self.meta = meta

    @property
    def meta(self):
        if hasattr(self, '_meta'):
            return self._meta

        #log.error('SqliteCache getter not implemented yet.')

    @meta.setter
    def meta(self, value):
        """ set meta """
        #log.error('SqliteCache setter not implemented yet. Should probably be done in bulk anyway ...')


SqliteCache._bind_flavours()


class SymlinkCache(CachePath):

    def __init__(self, *args, meta=None, **kwargs):
        if meta is not None:
            self.meta = meta

    @property
    def meta(self):
        if hasattr(self, '_meta'):
            return self._meta

        if self.is_symlink():
            if not self.exists():  # if a symlink exists it is something other than what we want
                #assert pathlib.PurePosixPath(self.name) == self.readlink().parent.parent
                return PathMeta.from_symlink(self)
            else:
                raise exc.PathExistsError(f'Target of symlink exists!\n{self} -> {self.resolve()}')

        else:
            return super().meta

    @meta.setter
    def meta(self, pathmeta):
        if not self.exists():
            # if the path does not exist write even temporary to disk
            if self.is_symlink():
                meta = self.meta
                if meta == pathmeta:
                    log.debug(f'Metadata unchanged for {meta.id}. Not updating.')
                    return

                if meta.id != pathmeta.id:
                    msg = ('Existing cache id does not match new id!\n'
                           f'{self!r}\n'
                           f'{meta.id} != {pathmeta.id}\n'
                           f'{meta.as_pretty()}\n'
                           f'{pathmeta.as_pretty()}')
                    log.critical(msg)
                    meta_newer = 'Meta newer. Not updating.'
                    pathmeta_newer = 'Other meta newer.'
                    msg = '{}'  # apparently I was out of my mind when I wrote this originally ...
                    if meta.updated is None and pathmeta.updated is None:
                        log.warning ('no change since either has an updated value (wat)')
                        return #FIXME

                    if meta.updated > pathmeta.updated:
                        log.info(msg.format(meta_newer))
                        return  # this is the right thing to do for a sane filesystem
                    elif meta.updated < pathmeta.updated:
                        log.info(msg.format(pathmeta_newer))
                        # THIS IS EXPLICITLY ALLOWED
                    else:  # they are equal
                        extra = 'Both updated at the same time '
                        if meta.created is not None and pathmeta.created is not None:
                            if meta.created > pathmeta.created:
                                log.info(msg.format(extra + meta_newer))
                                return
                            elif meta.created < pathmeta.created:
                                log.info(msg.format(extra + pathmeta_newer))
                                # THIS IS EXPLICITLY ALLOWED
                            else:  # same created
                                log.info(msg.format('Identical timestamps. Not updating.'))
                                return
                        elif meta.created is not None:
                            log.info(msg.format(extra + 'Meta has datetime other does not. Not updating.'))
                            return
                        elif pathmeta.created is not None:
                            msg = msg.format(extra + 'Meta has no datetime other does.')
                            log.info(msg)
                            raise exc.MetadataIdMismatchError(msg)
                        else:  # both none
                            log.info(msg.format(extra + ('Identical update time both missing created time. '
                                                         'Not updating.')))
                            return
                    # equality
                # id mismatch all cases above should return or raise except for other metadata newer

                if meta.size is not None and pathmeta.size is None:
                    log.error('new meta has no size so will not overwrite')
                    return

                # FIXME do the timestamp dance above here
                log.debug('Metadata exists, but ids match so will update')

                # trash old versions instead of just unlinking
                pc = self.local.cache
                trash = pc.trash
                self.rename(trash / f'{pc.parent.id}-{meta.id}-{self.name}')
                #self.unlink()

            # FIXME if an id starts with / then the local name is overwritten due to pathlib logic
            # we need to error if that happens
            #symlink = pathlib.PurePosixPath(self.local.name, pathmeta.as_symlink().as_posix().strip('/'))
            symlink = pathlib.PurePosixPath(self.local.name) / pathmeta.as_symlink()
            self.local.symlink_to(symlink)

        else:
            raise exc.PathExistsError(f'Path exists {self}')


SymlinkCache._bind_flavours()


class PrimaryCache(CachePath):
    @property
    def meta(self):
        #if hasattr(self, '_in_bootstrap'):
        #if hasattr(self, '_meta'):  # if we have in memory we are bootstrapping so don't fiddle about
            #return self._meta

        exists = self.exists()
        if exists:
            #log.debug(self)  # TODO this still gets hit a lot in threes
            meta = super().meta
            if meta:  # implicit else failover to backup cache
                return meta

        elif not exists and self._not_exists_cache and self.is_symlink():
            try:
                cache = self._not_exists_cache(self)
                return cache.meta
            except exc.NoCachedMetadataError as e:
                log.warning(e)

        if self._backup_cache:
            try:
                cache = self._backup_cache(self)
                meta = cache.meta
                if meta:
                    log.info(f'restoring from backup {meta}')
                    self._meta_setter(meta)  # repopulate primary cache from backup
                    return meta

            except exc.NoCachedMetadataError as e:
                log.warning(e)

    def _meta_setter(self, pathmeta, memory_only=False):
        """ we need memory_only for bootstrap I think """
        if not pathmeta:
            log.warning(f'Trying to set empty pathmeta on {self}')
            return

        if self.exists_not_symlink():  # if a file already exists just follow instructions
            super()._meta_setter(pathmeta)

        else:
            if not hasattr(self, '_remote') or self._remote is None:
                self._bootstrapping_id = pathmeta.id

            # need to run this to create directories
            self._bootstrap_prepare_filesystem(parents=False,
                                               fetch_data=False,
                                               size_limit_mb=0)

            if self.exists():  # we a directory now
                super()._meta_setter(pathmeta)

            elif self._not_exists_cache:
                cache = self._not_exists_cache(self, meta=pathmeta)

        if self._backup_cache:
            cache = self._backup_cache(self, meta=pathmeta)

        if hasattr(self, '_meta'):
            delattr(self, '_meta')

        if hasattr(self, '_id'):
            delattr(self, '_id')

    @staticmethod
    def _update_meta(old, new):
        if not old:
            return False, new  # if there is no file it is both different and not different

        if not new:
            return False, old

        file_is_different = False

        kwargs = {k:v for k, v in old.items()}
        if old.id != new.id:
            kwargs['old_id'] = old.id

        for k, vnew in new.items():
            vold = kwargs[k]

            if vnew is None or hasattr(vnew, '__iter__') and not vnew:
                # don't update with None or empty iterables
                continue

            if vold is not None and vold != vnew:
                log.info(f'{old.id} field {k} changed from {vold} -> {vnew}')
                if k in ('created', 'updated', 'size', 'checksum', 'file_id'):
                    file_is_different = True

            kwargs[k] = vnew

        if file_is_different:
            # strip fields missing from new in the case where
            # we aren't merging metadata from two different sources

            for k, vnew in new.items():
                if k == 'old_id':
                    continue

                if vnew is None:
                    log.debug(kwargs.pop(k))

        #old.updated == new.updated
        #old.updated < new.updated
        #old.updated > new.updated

        #old.created == new.created
        #old.created < new.created
        #old.created > new.created

        return file_is_different, PathMeta(**kwargs)

    def _meta_updater(self, pathmeta, fetch=True):
        original = self.meta
        file_is_different, updated = self._update_meta(original, pathmeta)
        # FIXME missing checksum is one source of problems here
        must_fetch = file_is_different and self.is_file() and self.exists() and fetch

        if must_fetch:
            try:
                # FIXME performance, and pathmeta.checksum is None case
                if self.local.content_different() and self.local.meta.checksum != pathmeta.checksum:
                    raise exc.LocalChangesError(f'not fetching {self}')

            except exc.NoRemoteFileWithThatIdError as e:
                log.warning('cant fetch remote file there may be untracked local changes for\n{self}')

            log.info(f'crumpling to preserve existing metadata\n{self}')
            trashed = self.crumple()

        try:
            self._meta_setter(updated)
            if must_fetch:
                self.fetch(size_limit_mb=None)

        except BaseException as e:
            log.error(e)
            if must_fetch:
                trashed.rename(self)
            raise e

        return file_is_different

    @property
    def children(self):
        """ direct children """
        # if you want the local children go on local
        # this will give us the remote children in the local context
        # going in the reverse direction with parents
        # we don't do because the parents here are already defined
        # if a file has moved on the remote we can detect that and error for now
        for child_remote in self.remote.children:
            child_cache = self / child_remote
            yield child_cache

    @property
    def rchildren(self):
        # FIXME cached rchildren vs local (working tree) rchildren vs remote rchildren
        # have to express the generator to build the index
        # otherwise child.parents will not work correctly (annoying)
        for child_remote in self.remote.rchildren:
            # FIXME is this causing the creation of folders during bootstrap ???
            args = child_remote._parts_relative_to(self.remote)  # usually this would just be one level
            child_cache = self._make_child(args, child_remote)
            #child_cache = self
            #child_path = self.__class__(self, *args)
            #child_path.remote = child

            yield child_cache

        # if organization
        # if dataset (usually going to be the fastest in most cases)
        # if collection (can end up very slow)
        # if package/file


PrimaryCache._bind_flavours()


class SshCache(PrimaryCache, EatCache):
    xattr_prefix = 'ssh'
    _backup_cache = SqliteCache
    _not_exists_cache = SymlinkCache
    cypher = default_cypher

    # FIXME probably need a way to dispatch to multiple remotes
    # based on that prefix, since the id actually determines the
    # remote class in this case since the class maintains the session
    # after init, the hostname is materialized into the id to disambiguate
    # the local path specifications

    checksum = AugmentedPath.checksum

    @property
    def anchor(self):
        if not hasattr(self, '_anchor') or self._anchor is None:
            raise ValueError('Cache anchor is none! Did you call '
                             'localpath.cache_init(id, anchor=True)?')

        return self._anchor

    @property
    def data(self):
        # there is no middle man for ssh so we go directly
        yield from self.remote.data


SshCache._bind_flavours()


# assign defaults

from augpathlib.core import LocalPath
SshCache._local_class = LocalPath



import git


class _Repo(git.Repo):  # FIXME should we subclass Repo for this or patch ??
    """ monkey patching """

    def getRef(self, ref_name):
        for ref in self.refs:
            if ref.name == ref_name:
                return ref

        else:
            raise ValueError(f'No ref with name: {ref_name}')


# monkey patch git.Repo
git.Repo.getRef = _Repo.getRef


class _Reference(git.Reference):
    """ monkey patching """
    def __enter__(self):
        """ Checkout the ref for this head.
        `git stash --all` beforehand and restore during __exit__.

        If the ref is the same, then the stash step still happens.
        If you need to modify the uncommitted state of a repo this
        is not the tool you should use. """

        if not self.is_valid():
            raise exc.InvalidRefError(f'Not a valid ref: {self.name}')

        self.__original_branch = self.repo.active_branch
        self.__stash = self.repo.git.stash('--all')  # always stash
        if self.__stash == 'No local changes to save':
            self.__stash = None

        if self == self.__original_branch:
            return self

        self.checkout()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _stash = self.repo.git.stash('--all')  # always stash on the way out as well
        if _stash == 'No local changes to save':
            stash = 'stash@{0}'
        else:
            stash = "stash@{1}"

        if self.__original_branch != self:
            self.__original_branch.checkout()

        # TODO check to make sure no other stashes were pushed on top
        if self.__stash is not None:
            self.repo.git.stash('pop', stash)

        self.__stash = None


# monkey patch git.Reference
git.Reference.__enter__ = _Reference.__enter__
git.Reference.__exit__ = _Reference.__exit__



import unittest
from augpathlib.utils import FileSize

class TestFileSize(unittest.TestCase):
    def test_0_getattr(self):
        s = FileSize(1000000000000000)
        getattr(s, 'hr')

    def test_1_str(self):
        s = FileSize(1)
        ss = str(s)

    def test_2_repr(self):
        s = FileSize(1)
        sr = str(s)



# -*- coding: utf-8 -*-
'''
Copyright © 2015, Robin David - MIT-Licensed
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
The Software is provided "as is", without warranty of any kind, express or implied, including but not limited
to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall
the authors or copyright holders X be liable for any claim, damages or other liability, whether in an action
of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other
dealings in the Software.
Except as contained in this notice, the name of the Robin David shall not be used in advertising or otherwise
to promote the sale, use or other dealings in this Software without prior written authorization from the Robin David.
'''
from ctypes import *

import sys, os
kernel32 = windll.kernel32

LPSTR     = c_wchar_p
DWORD     = c_ulong
LONG      = c_ulong
WCHAR     = c_wchar * 296
LONGLONG  = c_longlong

class LARGE_INTEGER_UNION(Structure):
    _fields_ = [
        ("LowPart", DWORD),
        ("HighPart", LONG),]

class LARGE_INTEGER(Union):
    _fields_ = [
        ("large1", LARGE_INTEGER_UNION),
        ("large2", LARGE_INTEGER_UNION),
        ("QuadPart",    LONGLONG),
    ]

class WIN32_FIND_STREAM_DATA(Structure):
    _fields_ = [
        ("StreamSize", LARGE_INTEGER),
        ("cStreamName", WCHAR),
    ]
    '''
    typedef struct _WIN32_FIND_STREAM_DATA {
      LARGE_INTEGER StreamSize;
      WCHAR         cStreamName[MAX_PATH + 36];
    } WIN32_FIND_STREAM_DATA, *PWIN32_FIND_STREAM_DATA;
    '''

class ADS():
    def __init__(self, filename):
        self.filename = filename
        self.streams = self.init_streams()

    def init_streams(self):
        file_infos = WIN32_FIND_STREAM_DATA()
        streamlist = list()

        findFirstStreamW = kernel32.FindFirstStreamW
        findFirstStreamW.restype = c_void_p

        myhandler = kernel32.FindFirstStreamW (LPSTR(self.filename), 0, byref(file_infos), 0)
        '''
        HANDLE WINAPI FindFirstStreamW(
          __in        LPCWSTR lpFileName,
          __in        STREAM_INFO_LEVELS InfoLevel, (0 standard, 1 max infos)
          __out       LPVOID lpFindStreamData, (return information about file in a WIN32_FIND_STREAM_DATA if 0 is given in infos_level
          __reserved  DWORD dwFlags (Reserved for future use. This parameter must be zero.) cf: doc
        );
        https://msdn.microsoft.com/en-us/library/aa364424(v=vs.85).aspx
        '''
        p = c_void_p(myhandler)

        if file_infos.cStreamName:
            streamname = file_infos.cStreamName.split(":")[1]
            if streamname: streamlist.append(streamname)

            while kernel32.FindNextStreamW(p, byref(file_infos)):
                streamlist.append(file_infos.cStreamName.split(":")[1])

        kernel32.FindClose(p)  # Close the handle

        return streamlist

    def __iter__(self):
        return iter(self.streams)

    def has_streams(self):
        return len(self.streams) > 0

    def full_filename(self, stream):
        return "%s:%s" % (self.filename, stream)

    def add_stream_from_file(self, filename):
        if os.path.exists(filename):
            with open(filename, "rb") as f:
                content = f.read()
            return self.add_stream_from_string(filename, content)
        else:
            print("Could not find file: {0}".format(filename))
            return False

    def add_stream_from_string(self, stream_name, string):
        fullname = self.full_filename(os.path.basename(stream_name))
        if os.path.exists(fullname):
            print("Stream name already exists")
            return False
        else:
            fd = open(fullname, "wb")
            fd.write(string)
            fd.close()
            self.streams.append(stream_name)
            return True

    def delete_stream(self, stream):
        try:
            os.remove(self.full_filename(stream))
            self.streams.remove(stream)
            return True
        except:
            return False

    def get_stream_content(self, stream):
        fd = open(self.full_filename(stream), "rb")
        content = fd.read()
        fd.close()
        return content



import os
import atexit
import shutil
import pathlib
from tempfile import gettempdir
import pytest
import augpathlib as aug
from augpathlib import exceptions as exc
from augpathlib import LocalPath
from augpathlib import PrimaryCache, RemotePath
from augpathlib import EatCache, SymlinkCache
from augpathlib import PathMeta
from augpathlib.utils import onerror_windows_readwrite_remove

aug.utils.log.setLevel('DEBUG')
log = aug.utils.log.getChild('test')

onerror = onerror_windows_readwrite_remove if os.name == 'nt' else None

_pid = os.getpid()
this_file = LocalPath(__file__)
temp_path = aug.AugmentedPath(gettempdir(), f'.augpathlib-testing-base-{_pid}')
project_path = this_file.parent / 'test_local/test_project'

# insurance in case some test forgets to clean up after itself
atexit.register(lambda : (shutil.rmtree(temp_path, onerror=onerror)
                          if temp_path.exists() else None))

SKIP_NETWORK = ('SKIP_NETWORK' in os.environ or
                'FEATURES' in os.environ and 'network-sandbox' in os.environ['FEATURES'])
skipif_no_net = pytest.mark.skipif(SKIP_NETWORK, reason='Skipping due to network requirement')


class LocalPathTest(LocalPath):
    def metaAtTime(self, time):
        # we are cheating in order to do this
        return PathMeta(id=self._cache_class._remote_class.invAtTime(self, time))


LocalPathTest._bind_flavours()


test_base = LocalPathTest(__file__).parent / f'test-base-{_pid}'
test_path = test_base / 'test-container'


class CachePathTest(PrimaryCache, EatCache):
    xattr_prefix = 'test'
    #_backup_cache = SqliteCache
    _not_exists_cache = SymlinkCache


CachePathTest._bind_flavours()


class RemotePathTest(RemotePath):
    anchor = test_path
    ids = {0: anchor}  # time invariant
    dirs = {2, 3, 4, 8, 9, 11, 12, 13, 14, 16, 17, 18}
    index_at_time = {1: {1: anchor / 'a.e',

                         2: anchor / 'c.e',

                         3: anchor / 'ee',
                         4: anchor / 'ee/ff',
                         5: anchor / 'ee/ff/gg',

                         8: anchor / 'ii',
                         9: anchor / 'ii/jj',
                         10: anchor / 'ii/jj/kk',

                         13: anchor / 'nn',
                         14: anchor / 'nn/oo',
                         15: anchor / 'nn/oo/pp',

                         18: anchor / 't.e',},
                     2: {1: anchor / 'b.e',

                         2: anchor / 'd.e',

                         3: anchor / 'hh/',
                         4: anchor / 'hh/ff',
                         5: anchor / 'hh/ff/gg',

                         11: anchor / 'll',
                         12: anchor / 'll/mm',
                         10: anchor / 'll/mm/kk',

                         16: anchor / 'qq',
                         17: anchor / 'qq/rr',
                         15: anchor / 'qq/rr/ss',

                         18: anchor / 't.e',

                         19: anchor / 'u.e',}}

    for ind in index_at_time:
        index_at_time[ind].update(ids)

    test_time = 2

    def __init__(self, thing_with_id, cache=None):
        if isinstance(thing_with_id, int):
            thing_with_id = str(thing_with_id)

        super().__init__(thing_with_id, cache)
        self._errors = []

    def is_dir(self):
        return int(self.id) in self.dirs

    def is_file(self):
        return not self.is_dir()

    def as_path(self):
        return pathlib.PurePosixPath(self.index_at_time[self.test_time][int(self.id)].relative_to(self.anchor))

    @classmethod
    def invAtTime(cls, path, index):
        path = cls.anchor / path
        return str({p:i for i, p in cls.index_at_time[index].items()}[path])

    @property
    def name(self):
        return self.as_path().name

    @property
    def parent(self):
        if int(self.id) == 0:
            return None

        rlu = self.as_path().parent
        return self.__class__(self.invAtTime(rlu, self.test_time))

    @property
    def meta(self):
        return PathMeta(id=self.id)

    def __repr__(self):
        p = self.as_path()
        return f'{self.__class__.__name__} <{self.id!r} {p!r}>'


# set up cache hierarchy
LocalPathTest._cache_class = CachePathTest
CachePathTest._local_class = LocalPathTest
CachePathTest._remote_class = RemotePathTest
RemotePathTest._cache_class = CachePathTest

# set up testing anchor (must come after the hierarchy)
CachePathTest.anchor = test_path
# note: the this creates a symlink which the rests of the tests expect
if test_path.exists():
    test_path.rmtree(onerror=onerror)
CachePathTest.anchor = CachePathTest(test_path, meta=PathMeta(id='0'))
test_path.unlink()


class TestPathHelper:
    @classmethod
    def setUpClass(cls):
        if cls.test_base.exists():
            shutil.rmtree(cls.test_base, onerror=onerror)

        cls.test_base.mkdir()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.test_base, onerror=onerror)

    def setUp(self, init_cache=True):
        if self.test_path.exists():  # in case something went wrong with a previous test
            shutil.rmtree(self.test_path, onerror=onerror)

        self.test_path.mkdir()
        if init_cache:
            self.test_path.cache_init('0')

    def tearDown(self):
        shutil.rmtree(self.test_path, onerror=onerror)


TestPathHelper.test_base = test_base  # there are here to prevent
TestPathHelper.test_path = test_path  # pytest from discovering them



"""
Classes for storing and converting metadata associated with path like objects.
"""
import pickle
import struct
import pathlib
from datetime import datetime
from dateutil import parser as dateparser
from terminaltables import AsciiTable
from augpathlib import exceptions as exc
from augpathlib.utils import log, FileSize


def isoformat(datetime_instance, timespec='auto'):
    return (datetime_instance
            .isoformat(timespec=timespec)
            .replace('.', ',')
            .replace('+00:00', 'Z'))


class PathMeta:
    """ Internal representation that every other format converts through. """

    # TODO register xattr prefixes

    @classmethod
    def from_metastore(cls, blob, prefix=None):
        """ db entry """
        xattrs = pickle.loads(blob)
        return cls.from_xattrs(xattrs, prefix)


    def __init__(self,
                 size=None,
                 created=None,
                 updated=None,
                 checksum=None,
                 etag=None,
                 chunksize=None,  # used for properly checksumming?
                 id=None,
                 file_id=None,
                 old_id=None,
                 gid=None,  # needed to determine local writability
                 user_id=None,
                 mode=None,
                 errors=tuple(),
                 **kwargs):

        if not file_id and file_id is not None and file_id is not 0:
            raise TypeError('wat')
        if created is not None and not isinstance(created, int) and not isinstance(created, datetime):
            _created = created
        else:
            self._created_ok = created
            _created = None

        if updated is not None and not isinstance(updated, int) and not isinstance(updated, datetime):
            _updated = updated
        else:
            self._updated_ok = updated
            _updated = None

        if id is not None and not isinstance(id, str):
            # no implicit type mutation, the system providing the ids
            # is where the information about how to do the converstion lives
            # we don't handle it here
            raise TypeError(f'id must be a string! {id!r}')

        self.size = size if size is None else FileSize(size)
        self._created = _created
        self._updated = _updated
        self.checksum = checksum
        self.etag = etag
        self.chunksize = chunksize
        self.id = id
        self.file_id = file_id
        self.old_id = old_id
        self.gid = gid
        self.user_id = user_id
        self.mode = mode
        self.errors = tuple(errors) if errors else tuple()
        if kwargs:
            log.warning(f'Unexpected meta values! {kwargs}')
            self.__kwargs = kwargs  # roundtrip values we don't explicitly handle

    #def as_xattrs(self, prefix=None):
        #log.debug(f'{self} {prefix}')
        #embed()
        #return self._as_xattrs(self, prefix=prefix)

    @property
    def created(self):
        if not hasattr(self, '_created_ok'):
            self._created_ok = dateparser.parse(self._created)

        return self._created_ok

    @property
    def updated(self):
        if not hasattr(self, '_updated_ok'):
            self._updated_ok = dateparser.parse(self._updated)

        return self._updated_ok

    def content_different(self, other):
        """ is there any evidence that the file itself changed? """
        # branch on checksum presence since it is definitional
        if self.checksum and other.checksum:
            return self.checksum != other.checksum
        else:
            return (self.size != other.size or
                    self.updated < other.updated)

    def items(self):
        out = [(k, v) for k, v in self.__dict__.items() if k[0] != '_']  # FIXME nonfields?
        return out + [('created', self.created), ('updated', self.updated)]
        #for field in self.fields:
            #yield field, getattr(self, field)

    def as_metastore(self, prefix=None):
        # FIXME prefix= is a bad api ...
        """ db entry """  # TODO json blob in sqlite? can it index?
        return pickle.dumps(self.as_xattrs(prefix))

    def keys(self):
        yield from self

    def __getitem__(self, key):
        try:
            getattr(self, key)
        except AttributeError as e:
            raise KeyError(key)

    def __iter__(self):
        yield from (k for k in self.__dict__ if not k.startswith('_'))

    def __reduce__(self):
        return (self.__class__, {k:v for k, v in self.items() if v is not None})

    def __repr__(self):
        _dict = {k:v for k, v in self.items() if not k.startswith('_')}
        return f'{self.__class__.__name__}({_dict})'

    def __eq__(self, other):
        if isinstance(other, PathMeta):
             for field, value in self.__dict__.items():
                  if value != getattr(other, field):
                       return False
             else:
                  return True

    def __bool__(self):
        for k, v in self.__dict__.items():
            if v is not None:
                if k == 'errors' and not v:  # empty tuple ok
                    continue

                return True
        else:
            return False


class _PathMetaConverter:
    """ Base class for implementing PathMeta format converts.

        I haven't figured out how to properly abstract registering
        as_format_name and from_format_name yet, so look at one of
        the __init__ methods below to see how it is done. """

    pathmetaclass = PathMeta
    format_name = None


class _PathMetaAsSymlink(_PathMetaConverter):
    """ Convert to and from symlinks to nowhere.
        You can prepend whatever other info you want in front of this """

    format_name = 'symlink'
    empty = '#'
    fieldsep = '.'  # must match the file extension splitter for Path ...
    subfieldsep = ';'  # only one level, not going recursive in a filename ...
    versions = {'meta':('file_id',
                        'size',
                        'created',
                        'updated',
                        'checksum',
                        'old_id',
                        'gid',
                        'user_id',
                        'mode',
                        'errors',),
                'mdv1':('file_id',
                        'size',
                        'created',
                        'updated',
                        'checksum',
                        'chunksize',  # added
                        'old_id',
                        'gid',
                        'user_id',
                        'mode',
                        'errors',),
                'mdv2':('file_id',
                        'size',
                        'created',
                        'updated',
                        'checksum',
                        'etag',  # added
                        'chunksize',
                        'old_id',
                        'gid',
                        'user_id',
                        'mode',
                        'errors',),
    }
    write_version = 'mdv2'  # update this on new version
    extras = 'size.hr',

    def __init__(self):
        # register functionality on PathMeta
        def as_symlink(self, _as_symlink=self.as_symlink):
            return _as_symlink(self)

        @classmethod
        def from_symlink(cls, symlink_path, _from_symlink=self.from_symlink):
            return _from_symlink(symlink_path)

        self.pathmetaclass.as_symlink = as_symlink
        self.pathmetaclass.from_symlink = from_symlink

    @property
    def order_all(self):
        return self.versions[self.write_version] + self.extras

    def encode(self, field, value):
        if field == 'file_id':
            if not value:
                if value is not None:
                    log.critical(f'{value!r} for file_id empty but not None!')
                value = None

        if value is None:
            return self.empty

        if field in ('errors',):
            return self.subfieldsep.join(value)

        if field == 'checksum':
            return value.hex()  # raw hex may contain field separators :/

        return _str_encode(field, value)

    def decode(self, field, value):
        if value == self.empty:
            return None

        if field == 'errors':
            return [_ for _ in value.split(self.subfieldsep) if _]

        elif field in ('created', 'updated'):
            setattr(self, '_' + field, value)
            return value

        elif field == 'checksum':  # FIXME checksum encoding ...
            #return value.encode()
            return bytes.fromhex(value)

        elif field == 'etag':
            checksum, strcount = value.rsplit('-', 1)
            count = int(strcount)
            return bytes.fromhex(checksum), count

        elif field == 'user_id':
            try:
                return int(value)
            except ValueError:  # FIXME :/ uid vs owner_id etc ...
                return value

        elif field in ('id', 'mode', 'old_id'):
            return value

        else:
            try:
                return int(value)
            except ValueError as e:
                breakpoint()
                raise e

        return value

    def as_symlink(self, pathmeta):
        """ encode meta as a relative path to be appended as a child,
            not a sibbling """

        __ignoreme = object()

        def multigetattr(object, attr, default=__ignoreme):
            rest = None
            if '.' in attr:
                first, attr = attr.split('.')
                object = multigetattr(object, first, default=default)

            if default != __ignoreme:
                return getattr(object, attr, default)
            else:
                return getattr(object, attr)


        gen = (self.encode(field, multigetattr(pathmeta, field, None))
               for field in self.order_all)

        id = self.encode('id', pathmeta.id)

        path_string = id + f'/.{self.write_version}.' + self.fieldsep.join(gen)
        return pathlib.PurePosixPath(path_string)

    def from_symlink(self, symlink_path):
        """ contextual portion to make sure something weird isn't going on
            e.g. that a link got switched to point to another name somehow """
        raw_symlink = symlink_path.readlink(raw=True)
        pure_symlink = pathlib.PurePosixPath(raw_symlink)
        name, *parts = pure_symlink.parts
        msg = (symlink_path.name, name)
        assert symlink_path.name == name, msg
        return self.from_parts(parts)

    def from_parts(self, parts):
        data = parts[-1]
        _, version, *suffixes = data.split(self.fieldsep)
        order = self.versions[version]
        kwargs = {field:self.decode(field, value)
                  for field, value in zip(order, suffixes)}
        path = pathlib.PurePosixPath(*parts)
        kwargs['id'] = str(path.parent)
        return self.pathmetaclass(**kwargs)


class _PathMetaAsXattrs(_PathMetaConverter):
    """ Convert to and from unix xattrs. """

    format_name = 'xattrs'

    fields = ('size',
              'created',
              'updated',
              'checksum',
              'etag',
              'chunksize',
              'id',
              'file_id',
              'old_id',
              'gid',
              'user_id',
              'mode',
              'errors')

    def __init__(self):
        # register functionality on PathMeta
        def as_xattrs(self, prefix=None, _as_xattrs=self.as_xattrs):
            #log.debug(f'{self} {prefix}')
            return _as_xattrs(self, prefix=prefix)

        # as a note: information hiding in python is ... weird ...
        # even when that isn't what you set out to do ...
        @classmethod
        def from_xattrs(cls, xattrs, prefix=None, path_object=None,
                        _from_xattrs=self.from_xattrs):
            # FIXME cls(**kwargs) vs calling self.pathmetaclass
            return _from_xattrs(xattrs, prefix=prefix, path_object=path_object)

        self.pathmetaclass.as_xattrs = as_xattrs
        self.pathmetaclass.from_xattrs = from_xattrs

    @staticmethod
    def deprefix(string, prefix):
        if string.startswith(prefix):
            string = string[len(prefix):]
            return string

    def from_xattrs(self, xattrs, prefix=None, path_object=None):
        """ decoding from bytes """
        _decode = getattr(path_object, 'decode_value', None)
        if path_object and _decode is not None:
            # some classes may need their own encoding rules _FOR NOW_
            # we will remove them once we standardize the xattrs format
            def decode(field, value, dv=_decode):
                out = dv(field, value)
                if value is not None and out is None:
                    out = self.decode(field, value)

                return out

        else:
            decode = self.decode

        if prefix:
            prefix += '.'
            kwargs = {k:decode(k, v)
                      for kraw, v in xattrs.items()
                      for k in (self.deprefix(kraw.decode(), prefix),)
                      # yes, we do have to filter out things that do
                      # not match the prefix because an unprefixed key
                      # of the same name would collide, also things with
                      # alt prefixes cause lots of warnings during init
                      if k is not None}
        else:  # ah manual optimization
            kwargs = {k:decode(k, v)
                      for kraw, v in xattrs.items()
                      for k in (kraw.decode(),)}

        return self.pathmetaclass(**kwargs)

    def as_xattrs(self, pathmeta, prefix=None):
        """ encoding to bytes """
        #log.debug(pathmeta)
        out = {}
        for field in self.fields:
            value = getattr(pathmeta, field)
            empty_iterable = hasattr(value, '__iter__') and not value
            if value is not None and not empty_iterable:  # xattrs discard type
                value_bytes = self.encode(field, value)
                if prefix:
                    key = prefix + '.' + field
                else:
                    key = field

                key_bytes = key.encode()
                out[key_bytes] = self.encode(field, value)

        return out

    def encode(self, field, value):
        #if field in ('created', 'updated') and not isinstance(value, datetime):
            #field.replace(cls.path_field_sep, ',')  # FIXME hack around iso8601
            # turns it 8601 isnt actually standard >_< with . instead of , sigh

        empty_iterable = hasattr(value, '__iter__') and not value
        if value is None or empty_iterable:
            raise TypeError('cannot encode an empty value')

        try:
            return _bytes_encode(field, value)
        except exc.UnhandledTypeError:
            log.warning(f'conversion not implemented for field {field}')

        raise exc.UnhandledTypeError(f'dont know what to do with {value!r}')

    def decode(self, field, value):
        if field in ('created', 'updated'):  # FIXME human readable vs integer
            try:
                # needed for legacy cases
                value, = struct.unpack('d', value)
                return datetime.fromtimestamp(value)
            except struct.error:
                pass
            vd = value.decode()
            setattr(self, '_' + field, vd)  # FIXME with timezone vs without ...
            return vd

        elif field == 'checksum':
            return value

        elif field == 'etag':
            # struct pack this sucker so the count can fit as well?
            value = value.decode()  # FIXME
            checksum, strcount = value.rsplit('-', 1)
            count = int(strcount)
            return bytes.fromhex(checksum), count

        elif field == 'errors':
            value = value.decode()
            return tuple(_ for _ in value.split(';') if _)

        elif field == 'user_id':
            try:
                return int(value)
            except ValueError:  # FIXME :/ uid vs owner_id etc ...
                return value.decode()

        elif field in ('id', 'mode', 'old_id'):
            return value.decode()

        elif field not in self.fields:
            log.warning(f'Unhandled field {field}')
            return value

        else:
            try:
                return int(value)
            except ValueError as e:
                log.exception(f'{field} {value}')
                raise e

class _PathMetaAsPretty(_PathMetaConverter):
    """ Convert to and from unix xattrs. """

    format_name = 'pretty'

    fields = ('size',
              'created',
              'updated',
              'checksum',
              'etag',
              'chunksize',
              'id',
              'file_id',
              'old_id',
              'gid',
              'user_id',
              'mode',
              'errors')

    def __init__(self):
        # register functionality on PathMeta
        def as_pretty(self, pathobject=None, title=None, human=False,
                      _as_pretty=self.as_pretty):
            return _as_pretty(self,
                              pathobject=pathobject,
                              title=title,
                              human=human)

        @classmethod
        def from_pretty(cls, pretty, _from_pretty=self.from_pretty):
            # FIXME cls(**kwargs) vs calling self.pathmetaclass
            return _from_pretty(pretty)

        self.pathmetaclass.as_pretty = as_pretty
        self.pathmetaclass.from_pretty = from_pretty

        self.maxf = max([len(f) for f in self.fields])

    def encode(self, field, value):
        if field == 'errors':
            return list(value)

        if field == 'checksum':
            if isinstance(value, bytes):
                value = value.hex()

        try:
            return _str_encode(field, value)
        except exc.UnhandledTypeError:
            log.warning(f'conversion not implemented for field {field}')
        
        return value

    def rows(self, meta, human=False):
        def key(kv):
            k, v = kv
            if k in self.fields:
                return 0, self.fields.index(k), k, v
            else:
                return 1, 0, k, v

        return sorted(([k, v.hr
                        if human and isinstance(v, FileSize)
                        else self.encode(k, v)]
                       for k, v in meta.items()
                       if (v is not None and
                           (isinstance(v, tuple) and
                            v or not isinstance(v, tuple)))),
                      key=key)

    def as_pretty(self, pathmeta, pathobject=None, title=None, human=False):
        if title is None:
            if pathobject is not None:
                title = pathobject.name
            else:
                title = ''
            
        h = [['Key', f'Value    {title}']]
        rows = h + self.rows(pathmeta, human=human)

        try:
            table = AsciiTable(rows, title=title).table
        except TypeError as e:
            breakpoint()
            raise e

        return table

    def from_pretty(self, pretty):
        raise NotImplementedError('yeah ... really don\'t want to do this')


class _PathMetaAsPrettyDiff(_PathMetaAsPretty):
    """ TODO """
    format_name = 'pretty-diff'

    def __init__(self):
        # register functionality on PathMeta
        def as_pretty_diff(self, othermeta, pathobject=None, title=None,
                           human=False, _as_pretty_diff=self.as_pretty_diff):
            return _as_pretty_diff(self,
                                   othermeta,
                                   pathobject=pathobject,
                                   title=title,
                                   human=human)

        @classmethod
        def from_pretty_diff(cls, pretty,
                             _from_pretty_diff=self.from_pretty_diff):
            # FIXME cls(**kwargs) vs calling self.pathmetaclass
            return _from_pretty_diff(pretty)

        self.pathmetaclass.as_pretty_diff = as_pretty_diff
        self.pathmetaclass.from_pretty_diff = from_pretty_diff

        self.maxf = max([len(f) for f in self.fields])

    def as_pretty_diff(self, pathmeta, othermeta, pathobject=None, title=None,
                       human=False):
        if title is None:
            if pathobject is not None:
                title = pathobject.relative_to(pathobject.cwd()).as_posix()
            else:
                title = ''
            
        if pathmeta.content_different(othermeta):
            title = red.format(title)

        def merge(a, b):
            keys = [k for k, _ in a]
            okeys = [ok for ok, _ in b]
            kind = [(k, okeys.index(k) - i, k)
                    if k in okeys else
                    (None, 0, k) for i, k in enumerate(keys)]
            okind = [(ok, keys.index(ok) - i, ok)
                     if ok in keys else
                     (None, 0, ok) for i, ok in enumerate(okeys)]
            long, short = ((kind, okind)
                           if len(kind) > len(okind) else
                           (okind, kind))
            start_align = None
            aligned = []
            offset = 0
            for i, (k, v, av) in enumerate(long):
                i -= offset
                if k is None and start_align is None:
                    aligned.append(av)
                    aligned.append(short[i + v][-1])
                elif k is not None:
                    if v <= 0:
                        start_align = i
                        aligned.append(av)
                    else:
                        for j in range(v):
                            aligned.append(short[i + j][-1])
                        offset += v
                        aligned.append(av)

                elif k is None:
                    aligned.append(av)
                # FIXME what to do if short has a key that long does not?

            lkv = len(a[0])
            lokv = len(b[0])
            kv = {k:[k, v] for k, v in a}
            okv = {k:[k, v] for k, v in b}
            for k in aligned:
                l = kv[k] if k in kv else ([''] * lkv)
                r = okv[k] if k in okv else ([''] * lokv)
                yield l + r

        h = [['Key', 'Value', 'Key Other', 'Value Other']]
        rows = h + list(merge(self.rows(pathmeta, human=human),
                              self.rows(othermeta, human=human)))
        try:
            table = AsciiTable(rows, title=title).table
        except TypeError as e:
            breakpoint()
            raise e

        return table

    def from_pretty_diff(self, pretty):
        raise NotImplementedError('yeah ... really don\'t want to do this')


class _EncodeByField:
    def __call__(self, field, value):
        if type(value) == str:
            return value

        try:
            return getattr(self, field)(value)
        except AttributeError as e:
            raise exc.UnhandledTypeError(field) from e
        
    def _datetime(self, value):
        if not isinstance(value, datetime):
            raise TypeError(f'{type(value)} is not a datetime for {value}')

        has_tz = (value.tzinfo is not None and
                  value.tzinfo.utcoffset(None) is not None)
        value = isoformat(value)
        if not has_tz:
            log.warning('why do you have a timestamp without a timezone ;_;')

        return value

    def _list(self, value):
        return ';'.join(value)

    def errors(self, value):
        return self._list(value)

    def created(self, value):
        return self._datetime(value)

    def updated(self, value):
        return self._datetime(value)

    def id(self, value):
        if not isinstance(value, str):
            raise TypeError('ids must always be strings!')

        return value

    def size(self, value): return str(value)
    def checksum(self, value): return value
    def etag(self, value): return f'{value[0].hex()}-{value[1]}'  # checksum-count  # TODO pack this?
    def chunksize(self, value): return str(value)
    def file_id(self, value): return str(value)
    def gid(self, value): return str(value)
    def user_id(self, value): return str(value)


class _EncodeByFieldBytes(_EncodeByField):
    def __call__(self, field, value):
        if isinstance(value, bytes):
            return value

        out = super().__call__(field, value)
        if isinstance(out, str):
            out = out.encode()

        return out


# load encoders
_str_encode = _EncodeByField()
_bytes_encode = _EncodeByFieldBytes()

# register helpers
_PathMetaAsSymlink()
_PathMetaAsXattrs()
_PathMetaAsPretty()
_PathMetaAsPrettyDiff()  # TODO



from augpathlib.meta import PathMeta
from augpathlib.core import (AugmentedPath,
                             AugmentedPathPosix,
                             AugmentedPathWindows,
                             XopenPath,
                             LocalPath,
                             EatPath)
from augpathlib.caches import (CachePath,
                               PrimaryCache,
                               SqliteCache,
                               SymlinkCache,
                               EatCache,
                               SshCache)
from augpathlib.remotes import RemotePath
from augpathlib.utils import StatResult, FileSize, etag
from augpathlib.zip import ZipHelper, ZipPath

try:
    from augpathlib.repo import RepoHelper, RepoPath
except ImportError as e:
    class RepoHelper:
        def __init__(self, *args, __error=e, **kwargs):
            raise ImportError(f'{self.__class__.__name__} could not be imported '
                              'due to a previous ImportError') from __error


    class RepoPath(RepoHelper):
        pass


__all__ = [
    'StatResult',
    'FileSize',
    'etag',

    'PathMeta',

    'AugmentedPath',
    'XattrPath',
    'RepoPath',
    'XopenPath',
    'LocalPath',

    'CachePath',
    'PrimaryCache',
    'SqliteCache',
    'SymlinkCache',
    'XattrCache',
    'SshCache',

    'RemotePath',
]

__version__ = '0.0.21'



import os
import unittest
from socket import gethostname
from pathlib import PurePath
import pytest
from augpathlib import LocalPath as Path
from augpathlib.caches import SshCache, ReflectiveCache
from augpathlib.remotes import SshRemote
from .common import project_path, TestPathHelper, skipif_no_net

@skipif_no_net
@pytest.mark.skipif(os.name == 'nt', reason='No pxssh for windows at the moment.')
@pytest.mark.skipif('CI' in os.environ, reason='Requires ssh/.config to be set up correctly.')
class TestSshRemote(TestPathHelper, unittest.TestCase):
    def setUp(self):
        super().setUp(init_cache=False)
        self.SshRemote = SshRemote._new(Path, SshCache)

        project_path = Path(self.test_path)
        self.project_path = project_path
        remote_root = PurePath(Path(__file__).parent)  # the 'remote' target
        remote_id = remote_root.as_posix()

        hostname = gethostname()
        # this_folder.meta is sort of one extra level of host keys
        try:
            anchor = project_path.cache_init(hostname + ':' + remote_id, anchor=True)
        except TypeError:  # pxssh fail
            anchor = project_path.cache_init(hostname + '-local:' + remote_id, anchor=True)

        # FIXME remote_root doesn't actually work for ssh remotes, it is always '/'
        #anchor = project_path.cache_init('/')  # this_folder.meta is sort of one extra level of host keys
        self.this_file = Path(__file__)
        self.this_file_darkly = self.SshRemote(__file__)
        tfd_cache = self.this_file_darkly.cache_init()

    def test_checksum(self):
        assert self.this_file_darkly.meta.checksum == self.this_file.meta.checksum

    def test_meta(self):
        #hrm = this_file_darkly.meta.__dict__, this_file_darkly.local.meta.__dict__
        #assert hrm[0] == hrm[1]
        rm = self.this_file_darkly.meta
        lm = Path(__file__).meta

        rmnid = {k:v for k, v in rm.items() if k != 'id'}
        lmnid = {k:v for k, v in lm.items() if k != 'id'}
        bads = []
        for k, rv in rmnid.items():
            lv = lmnid[k]
            if rv != lv:
                bads.append((lv, rv))

        assert not bads, bads

    def test_data(self):
        # FIXME this_file_darkly.local seems to not be in quite the right location?
        #assert list(self.this_file_darkly.data) == list(self.this_file_darkly.local.data)
        assert list(self.this_file_darkly.data) == list(self.this_file.data)

    #stats, checks = this_file_darkly.parent.children  # FIXME why does this list the home directory!?
    def test_access(self):
        f = self.SshRemote('/root/this-file-does-not-exist')
        assert not f.access('read')
        f = self.SshRemote(__file__)
        assert f.access('write')



import os
import sys
import errno
import shutil
import pathlib
import tempfile
import warnings
import mimetypes
import subprocess
from time import sleep
from errno import ELOOP, ENOENT, ENOTDIR, EBADF
from datetime import datetime, timezone
from functools import wraps
from itertools import chain
try:
    import magic  # from sys-apps/file consider python-magic ?
    _have_magic = True
except (AttributeError, ImportError, TypeError) as e:
    _have_magic = False

#import psutil  # import for experimental xopen functionality
#from Xlib.display import Display
#from Xlib import Xatom
import augpathlib as aug
from augpathlib import swap
from augpathlib import exceptions as exc
from augpathlib.meta import PathMeta
from augpathlib.utils import log, default_cypher, StatResult, etag
from augpathlib.utils import _bind_sysid_, AUG_XATTR_PREFIX

SPARSE_KEY = (AUG_XATTR_PREFIX + '.sparse')

_IGNORED_ERROS = (ENOENT, ENOTDIR, EBADF, ELOOP)
_IGNORED_WINERRORS = (
    123,  # 'The filename, directory name, or volume label syntax is incorrect' -> 22 EINVAL
)
if sys.platform == 'darwin':
    _IGNORED_ERROS += (errno.ENAMETOOLONG,)


if os.name != 'nt':
    import xattr
    XATTR_DEFAULT_NS = xattr.NS_USER
    def _ignore_error(exception):
        return (getattr(exception, 'errno', None) in _IGNORED_ERROS)

else:
    import winreg
    from augpathlib import pyads
    XATTR_DEFAULT_NS = 'user'
    def _ignore_error(exception):
        return ((getattr(exception, 'winerror', None) in _IGNORED_WINERRORS) or
                (getattr(exception, 'errno', None) in _IGNORED_ERROS))


if sys.version_info >= (3, 7):
    pathlib._IGNORED_ERROS += (ELOOP,)
    if sys.platform == 'darwin':
        # darwin gets very confused by self referential symlinks
        # and it seems that they stack up on eachother while being
        # dereferenced until the link name becomes too long
        pathlib._IGNORED_ERROS += (errno.ENAMETOOLONG,)

else:

    def _is_dir(entry):
        try:
            return entry.is_dir()
        except OSError as e:
            if not _ignore_error(e):
                raise

            return False

    def _iterate_directories(self, parent_path, is_dir, scandir):
        """ patched to fix is_dir() erron """
        yield parent_path
        try:
            entries = list(scandir(parent_path))
            for entry in entries:
                if _is_dir(entry) and not entry.is_symlink():
                    path = parent_path._make_child_relpath(entry.name)
                    for p in self._iterate_directories(path, is_dir, scandir):
                        yield p
        except PermissionError:
            return

    pathlib._RecursiveWildcardSelector._iterate_directories = _iterate_directories


def _catch_wrapper(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OSError as e:
            if not _ignore_error(e):
                raise

            return False

    return staticmethod(wrapped)


#pathlib._NormalAccessor.stat = _catch_wrapper(os.stat)  # can't wrap stat, pathlib needs the errors

# pathlib helpers, simplify the inheritance nightmare ...


class EatHelper:
    """ Extended attributes helper """

    @staticmethod
    def _base_helpers(pos_helpers, win_helpers):
        pos_helpers = tuple(set(pos_helpers + (XattrHelper,)))
        win_helpers = tuple(set(win_helpers + (ADSHelper,)))
        return pos_helpers, win_helpers

    @classmethod
    def _bind_flavours(cls, pos_helpers=tuple(), win_helpers=tuple()):
        super()._bind_flavours(*EatHelper._base_helpers(pos_helpers, win_helpers))


class ADSHelper(EatHelper):
    """ Windows NTFS equivalent of Xattrs is Alternate Data Streams
        This class allows ADS to pretend to work like xattrs.
    """

    _sparse_key = SPARSE_KEY

    @staticmethod
    def _key_convention(key, namespace):
        return namespace + '.' + key  # FIXME maybe include xattrs. as well ??

    def _stream(self, name):
        *start, last = self.parts
        if not start and len(last) == 1:
            # single letter file names with no extension
            # masquerade as drive letters on windows and
            # there seems to be nothing we can do about it
            raise ValueError('windows and single letter file names dont get along')

        return AugmentedPath(*start, last + ':' + name)

    @property
    def _streamname(self):
        *start, last = self.parts
        if ':' in last:
            return last.split(1, ':')[-1]

    @property
    def _streams(self):
        file_infos = pyads.WIN32_FIND_STREAM_DATA()
        #streamlist = list()

        findFirstStreamW = pyads.kernel32.FindFirstStreamW
        findFirstStreamW.restype = pyads.c_void_p

        myhandler = pyads.kernel32.FindFirstStreamW (pyads.LPSTR(self.as_posix()),
                                                     0,
                                                     pyads.byref(file_infos),
                                                     0)
        '''
        HANDLE WINAPI FindFirstStreamW(
        __in        LPCWSTR lpFileName,
        __in        STREAM_INFO_LEVELS InfoLevel, (0 standard, 1 max infos)
        __out       LPVOID lpFindStreamData, (return information about file in a WIN32_FIND_STREAM_DATA if 0 is given in infos_level
        __reserved  DWORD dwFlags (Reserved for future use. This parameter must be zero.) cf: doc
        );
        https://msdn.microsoft.com/en-us/library/aa364424(v=vs.85).aspx
        '''
        try:
            p = pyads.c_void_p(myhandler)

            if file_infos.cStreamName:
                streampath = file_infos.cStreamName
                if '::$DATA' not in streampath:
                    cleaned, _data_suffix = streampath.strip(':').rsplit(':$', 1)
                    yield self._stream(cleaned)

                while pyads.kernel32.FindNextStreamW(p, pyads.byref(file_infos)):
                    streampath = file_infos.cStreamName
                    if '::$DATA' not in streampath:
                        cleaned, _data_suffix = streampath.strip(':').rsplit(':$', 1)
                        yield self._stream(cleaned)

        finally:
            pyads.kernel32.FindClose(p)  # Close the handle

    def delxattr(self, key, fail=False, namespace=XATTR_DEFAULT_NS):
        name = self._key_convention(key, namespace)
        stream = self._stream(name)
        try:
            stream.unlink()
        except FileNotFoundError as e:
            if fail:
                raise exc.NoStreamError((self, key)) from e

    def setxattr(self, key, value, namespace=XATTR_DEFAULT_NS):
        if not isinstance(value, bytes):  # checksums
            raise TypeError(f'setxattr only accepts values already encoded to bytes!\n{value!r}')
        else:
            bytes_value = value

        if isinstance(key, bytes):
            key = key.decode()

        name = self._key_convention(key, namespace)
        stream = self._stream(name)
        log.debug(name)
        log.debug(stream)
        log.debug(bytes_value)
        with open(stream, 'wb') as f:
            f.write(bytes_value)

    def setxattrs(self, xattr_dict, namespace=XATTR_DEFAULT_NS):
        for k, v in xattr_dict.items():
            self.setxattr(k, v, namespace=namespace)

    def getxattr(self, key, namespace=XATTR_DEFAULT_NS):
        # we don't deal with types here, we just act as a dumb store
        name = self._key_convention(key, namespace)
        try:
            with open(self._stream(name), 'rb') as f:
                return f.read()

        except FileNotFoundError as e:
            raise exc.NoStreamError((self, key)) from e

    def _xattrs(self):
        out = {}
        for stream in self._streams:
            _base, k = stream.name.split(':', 1)
            with open(stream, 'rb') as f:
                v = f.read()

            out[k] = v

        return out

    def xattrs(self, namespace=XATTR_DEFAULT_NS):
        # decode keys later
        ns_length_p1 = len(namespace) + 1
        try:
            # we encode here to match the behavior of the posix version
            return {k[ns_length_p1:].encode():v for k, v in self._xattrs().items()
                    if k.startswith(namespace)}
        except FileNotFoundError as e:
            raise FileNotFoundError(self) from e


class XattrHelper(EatHelper):
    """ pathlib helper augmented with xattr support """

    _sparse_key = SPARSE_KEY.encode()

    def delxattr(self, key, fail=False, namespace=XATTR_DEFAULT_NS):
        try:
            xattr.remove(self.as_posix(), key, namespace=namespace)
        except OSError as e:
            if fail or e.errno != 61:  # 61 -> No data available
                raise e

    def setxattr(self, key, value, namespace=XATTR_DEFAULT_NS):
        if not isinstance(value, bytes):  # checksums
            raise TypeError('setxattr only accepts values already '
                            f'encoded to bytes!\n{value!r}')
        else:
            bytes_value = value

        xattr.set(self.as_posix(), key, bytes_value, namespace=namespace)

    def setxattrs(self, xattr_dict, namespace=XATTR_DEFAULT_NS):
        for k, v in xattr_dict.items():
            self.setxattr(k, v, namespace=namespace)

    def getxattr(self, key, namespace=XATTR_DEFAULT_NS):
        # we don't deal with types here, we just act as a dumb store
        try:
            return xattr.get(self.as_posix(), key, namespace=namespace)
        except OSError as e:
            if e.errno == errno.ENODATA or e.errno == errno.ENOATTR:
                raise exc.NoStreamError((self, key)) from e
            else:
                raise e

    def xattrs(self, namespace=XATTR_DEFAULT_NS):
        # decode keys later
        try:
            return {k:v for k, v in xattr.get_all(self.as_posix(), namespace=namespace)}
        except FileNotFoundError as e:
            raise FileNotFoundError(self) from e


# remote data about remote objects -> remote_meta
# local data about remote objects -> cache_meta
# local data about local objects -> meta
# remote data about local objects <- not relevant yet? or is this actually rr?

# TODO by convetion we could store 'children' of files in a .filename.ext folder ...
# obviously this doesn't always work
# and we would only want to do this for files that actually had annotations that
# needed a place to live ...


class AugmentedPath(pathlib.Path):
    """ extra conveniences, mostly things that are fixed in 3.7 using IGNORE_ERROS """

    _stack = []  # pushd and popd
    count = 0
    _debug = False  # sigh

    @classmethod
    def _bind_flavours(cls, pos_helpers=tuple(), win_helpers=tuple()):
        pos, win = cls._get_flavours()

        if pos is None:
            pos = type(f'{cls.__name__}Posix',
                       (*pos_helpers, cls, AugmentedPathPosix), {})

        if win is None:
            win = type(f'{cls.__name__}Windows',
                       (*win_helpers, cls, AugmentedPathWindows), {})

        cls.__abstractpath = cls
        cls.__posixpath = pos
        cls.__windowspath = win

    @classmethod
    def _get_flavours(cls):
        pos, win = None, None
        for subcls in cls.__subclasses__():  # direct only
            if subcls._flavour is pathlib._posix_flavour:
                pos = subcls
            elif subcls._flavour is pathlib._windows_flavour:
                win = subcls
            else:
                raise TypeError(f'unknown flavour for {cls} {cls._flavour}')

        return pos, win

    @classmethod
    def _abstract_class(cls):
        return cls.__abstractpath

    def __new__(cls, *args, **kwargs):
        if cls is cls.__abstractpath:
            cls = cls.__windowspath if os.name == 'nt' else cls.__posixpath
        self = cls._from_parts(args, init=False)
        if not self._flavour.is_supported:
            raise NotImplementedError("cannot instantiate %r on your system"
                                      % (cls.__name__,))
        self._init()
        return self

    def exists(self):
        """ Turns out that python doesn't know how to stat symlinks that point
            to their own children, which is fine because that is what we do
            so a reasonable way to short circuit the issue """
        try:
            return super().exists()
        except OSError as e:
            #log.error(e)   # too noisy ... though it reaveals we call exists a lot
            if not _ignore_error(e):
                raise

            return False

    def is_file(self):
        try:
            return super().is_file()
        except OSError as e:
            if not _ignore_error(e):
                raise

            return False

    def is_dir(self):
        try:
            return super().is_dir()
        except OSError as e:
            if not _ignore_error(e):
                raise

            return False

    def is_symlink(self):
        try:
            return super().is_symlink()
        except OSError as e:
            if not _ignore_error(e):
                raise

            return False

    def is_broken_symlink(self):
        """ The prime indicator that we have landed on a symlink that is being
            used to store data. The fullest indicator is the symlink loop if we
            want to implement a special checker that exploits errno.ELOOP 40. """
        return self.is_symlink() and not self.exists()

    def exists_not_symlink(self):
        return self.exists() and not self.is_symlink()

    def resolve(self):
        try:
            return super().resolve()
        except RuntimeError as e:
            msg = ('Unless this call to resolve was a mistake you should switch '
                   'to using readlink instead. Uncomment raise e to get a trace.\n'
                   'Alternately you might want to use absolute() in this situation instead?')
            raise RuntimeError(msg) from e

    def readlink(self, raw=False):
        """ this returns the string of the link only due to cycle issues """
        link = os.readlink(self)
        if isinstance(link, bytes):  # on pypy3 readlink still returns bytes
            link = link.decode()

        if raw:
            return link

        return pathlib.PurePath(link)

    def access(self, mode='read', follow_symlinks=True):
        """ types are 'read', 'write', and 'execute' """
        if mode in (os.R_OK, os.W_OK, os.X_OK):
            pass
        elif mode == 'read':
            mode = os.R_OK
        elif mode == 'write':
            mode = os.W_OK
        elif mode == 'execute':
            mode = os.X_OK
        else:
            raise TypeError(f'Unknown mode {mode}')

        # FIXME pypy3 stuck on 3.5 behavior
        return os.access(self.as_posix(), mode, follow_symlinks=follow_symlinks)

    def commonpath(self, other):
        return self.__class__(os.path.commonpath((self, other)))

    def to_relative(self, other):
        """ unfortunately pathlib's relative_to should
            actually be called relative_from because it is the
            relative path FROM the other path to the current path
            or rather it is relative_to_here_from_there which is
            not good because the final from is truncated, confusing
            everything also 'from here to there' is much more natural """

        rp = self.relative_to(other)
        return self.__class__(*['..' for _ in rp.parts[:-1]])

    def relative_path_to(self, other):  # from here
        """ an actual implementation of relative_to that works ... """
        return other.relative_path_from(self)

    def relative_path_from(self, other):  # to here
        """ relative path to self from the other path """
        base = self.commonpath(other)
        ort = other.to_relative(base)
        return ort / self.relative_to(base)

    def rename(self, target):
        os.rename(self, target)

    def swap_carefree(self, target):
        """ swap two paths, use atomic if available on the system
            and if the paths are on the same device otherwise fail

            side note: there doesn't seem to be a standard name for
            the superset of atomic and non-atomic, this operations
            is definitely not non-atomic, it is maybe-atomic, but
            maybe-atomic is too optimistic, thus carefree seems
            appropriately ... disinterested in the exact semantics """
        try:
            self.swap(target)
        except (Exception, NotImplementedError) as e:  # TODO clearer error handling
            temp_str = tempfile.mkdtemp(dir=target.parent)
            temp_dir = pathlib.Path(temp_str)
            temp = temp_dir / target.name
            # rename target -> temp
            # rename self -> target
            # rename target -> self
            target.rename(temp)
            self.rename(target)
            temp.rename(self)
            temp_dir.rmdir()

    if sys.platform != 'linux': # just look at us not using pathlib's infra ...
        _swap = swap.swap_not_implemented
    else:
        _swap = swap.swap_linux

    def swap(self, target):
        """ atomic swap of two paths

            this will raise and error in the following cases
            1. if the system has no atomic swap function such as renameat2
            2. if the paths to be swapped are on different devices
            3. if either path does not exist
        """
        se = self.exists()
        te = target.exists()
        if not se:
            msg = f'Both self and target must exist self does not! {self}'
            raise FileNotFoundError(msg)
        elif not te:
            msg = f'Both self and target must exist target does not! {target}'
            raise FileNotFoundError(msg)

        sd = self.stat().st_dev
        td = target.stat().st_dev
        if sd != td:
            msg = f'Self and target must be on the same device! {sd} != {td}'
            raise ValueError(msg) # FIXME find or make the correct error type

        self._swap(target)

    def rmtree(self, ignore_errors=False, onerror=None, DANGERZONE=False):
        """ DANGER ZONE """
        # FIXME make this atomic by renaming to a random name
        # that doesn't exist and then calling rmtree on that
        if not self.is_absolute():
            raise exc.WillNotRemovePathError(f'Only absolute paths can be removed recursively. {self}')

        if not (DANGERZONE is True):  # prevent python type coersion
            # TODO test in a chroot
            lenparts = len(self.parts)
            if lenparts <= 2:
                raise exc.WillNotRemovePathError(f'Will not remove top level paths. {self}')
            elif lenparts <= 3 and 'home' in self.parts:
                raise exc.WillNotRemovePathError(f'Will not remove home directories. {self}')
            elif self == pathlib.Path.cwd():
                raise exc.WillNotRemovePathError(f'Will not remove current working directory. {self}')

        try:
            if self.is_dir():
                if self.is_symlink():
                    # see the note in shutil.rmtree about race conditions
                    # for now we are going to hard fail if this case occures
                    raise OSError("Cannot call rmtree on a symbolic link")

                for path in self.iterdir():
                    if path.is_symlink():
                        # mimic shutil behavior and don't accidentally
                        # recurse through symlinks (keyword being curse)
                        path.unlink()
                    else:
                        path.rmtree(ignore_errors=ignore_errors,
                                    onerror=onerror,
                                    DANGERZONE=DANGERZONE)

                path = self
                self.rmdir()
            else:
                path = self
                self.unlink()

        except exc.WillNotRemovePathError:
            raise

        except BaseException as e:
            if not ignore_errors:
                if onerror is not None:
                    ftype = 'rmdir' if path.is_dir() else 'unlink'
                    func = getattr(path, ftype)
                    onerror(func, path, sys.exc_info())
                else:
                    raise e

    def chdir(self):
        os.chdir(self)

    def pushd(self):
        if self.is_dir():
            AugmentedPath._stack.append(self.cwd())
            self.chdir()
            print(*reversed(AugmentedPath._stack), self.cwd())
        else:
            raise NotADirectoryError(f'{self} is not a directory')

    @staticmethod
    def popd(N=0, n=False):
        """ see popd --help """
        # note that python lists append in the oppsite direction
        # so we invert the N dex
        reversed_index = - (N + 1)
        if AugmentedPath._stack:
            path = AugmentedPath._stack.pop(reversed_index)
            path.chdir()
            print(*reversed(AugmentedPath._stack), AugmentedPath.cwd())
            return path
        else:
            log.warning('popd: directory stack empty')

    def __enter__(self):
        if self.is_dir():
            self._entered_from = self.cwd()
            self.chdir()
            return self
        else:
            super().__enter__()

    def __exit__(self, t, v, tb):
        if hasattr(self, '_entered_from'):
            # if is_dir fails because of a change still try to return
            self._entered_from.chdir()
        else:
            super().__exit__(t, v, tb)

    @property
    def mimetype(self):
        if not self.is_absolute():  # needed for safe as_uri
            self = self.resolve()

        mime, encoding = mimetypes.guess_type(self.as_uri())
        if mime:
            return mime

    @property
    def encoding(self):
        if not self.is_absolute():  # needed for safe as_uri
            self = self.resolve()

        mime, encoding = mimetypes.guess_type(self.as_uri())
        if encoding:
            return encoding

    @property
    def _magic_mimetype(self):
        """ This can be slow because it has to open the files. """
        if self.exists():
            try:
                if hasattr(magic, 'detect_from_filename'):
                    # sys-apps/file python-magic api
                    return magic.detect_from_filename(self).mime_type
                else:
                    # python-magic
                    return magic.from_file(self.as_posix(), mime=True)
            except NameError as e:
                if not _have_magic:
                    msg = ('no module magic found from either python-magic '
                           'or from libmagic python bindings')
                    raise ModuleNotFoundError(msg) from e
                else:
                    raise e

    def checksum(self, cypher=default_cypher, extra_cyphers=tuple()):
        """ checksum() always recomputes from the data
            meta.checksum is static for cache and remote IF it exists """

        if self.is_file():
            if ((hasattr(self, '_cache_class') and
                 hasattr(self._cache_class, 'cypher') and
                 self._cache_class.cypher != cypher)):  # FIXME this could be static ...
                cypher = self._cache_class.cypher

            elif (hasattr(self, 'cypher') and
                  self.cypher != cypher):
                cypher = self.cypher

            m = cypher()
            extra = [c() for c in extra_cyphers]
            for chunk in self.data:
                m.update(chunk)
                for me in extra:
                    me.update(chunk)

            if extra_cyphers:
                return tuple(_.digest() for _ in (m, *extra))
            else:
                return m.digest()

    def copy_to(self, target, force=False):
        """ copy from a the current path object to a target path """
        if not target.exists() and not target.is_symlink() or force:
            # FIXME copytree does not happen, so sources cannot currently
            # be directories, what are the downsides of homogenizing that?
            shutil.copy2(self, target)

        else:
            raise exc.PathExistsError(f'{target}')

    def copy_from(self, source, force=False, copy_cache_meta=False):
        """ copy from a source path to the current path object """
        source.copy_to(self, force=force)

    def copy_outto(self, target, force=False):
        """ copy the current path out to a target directory """
        if not target.is_dir():
            raise NotADirectoryError(f'{target} is not a directory')

        target_file = target / self.name
        self.copy_to(target_file, force=force)
        return target_file

    def copy_infrom(self, source, force=False):
        """ copy into the current directory a file from somewhere else """
        return source.copy_outto(self, force=force)


class AugmentedPathPosix(AugmentedPath, pathlib.PosixPath): pass
class AugmentedPathWindows(AugmentedPath, pathlib.WindowsPath):
    _registry_drives = 'hklm', 'hkcu', 'HKLM', 'HKCU'

    if sys.version_info < (3, 8):
        # https://bugs.python.org/issue34384
        def readlink(self, raw=False):
            """ this returns the string of the link only due to cycle issues """
            link = os.readlink(str(self))
            if isinstance(link, bytes):  # on pypy3 readlink still returns bytes
                link = link.decode()

            if raw:
                return link

            return pathlib.PurePath(link)


def splitroot(self, part, sep='\\'):
    first = part[0:1]
    second = part[1:2]
    if (second == sep and first == sep):
        # XXX extended paths should also disable the collapsing of "."
        # components (according to MSDN docs).
        prefix, part = self._split_extended_path(part)
        first = part[0:1]
        second = part[1:2]
    else:
        prefix = ''
    third = part[2:3]
    if (second == sep and first == sep and third != sep):
        # is a UNC path:
        # vvvvvvvvvvvvvvvvvvvvv root
        # \\machine\mountpoint\directory\etc\...
        #            directory ^^^^^^^^^^^^^^
        index = part.find(sep, 2)
        if index != -1:
            index2 = part.find(sep, index + 1)
            # a UNC path can't have two slashes in a row
            # (after the initial two)
            if index2 != index + 1:
                if index2 == -1:
                    index2 = len(part)
                if prefix:
                    return prefix + part[1:index2], sep, part[index2+1:]
                else:
                    return part[:index2], sep, part[index2+1:]
    drv = root = ''
    if second == ':' and first in self.drive_letters:
        drv = part[:2]
        part = part[2:]
        first = third
    else:
        index1 = part.find(':')
        index2 = part.find(sep)
        if index1 != -1 and (index2 == -1 or index1 < index2):
            maybe_drv = part[:index1]
            if maybe_drv in self.drive_letters:
                drv = part[:index1]
                part = part[index1:]
                first = part[0:1]

    if first == sep:
        root = first
        part = part.lstrip(sep)
    return prefix + drv, root, part


pathlib._WindowsFlavour.drive_letters.update(AugmentedPathWindows._registry_drives)
pathlib._WindowsFlavour.splitroot = splitroot
pathlib._windows_flavour.splitroot = pathlib._WindowsFlavour().splitroot

AugmentedPath._bind_flavours()


class EatPath(EatHelper, AugmentedPath):

    # NOTE _sparse_key is set on each helper

    def _sparse_root(self):
        parent = self.parent
        if self != parent:
            try:
                return self.getxattr(self._sparse_key) == b'1'
            except exc.NoStreamError:
                return parent._sparse_root()

    def is_sparse(self):
        return self._sparse_root() is not None

    def _clear_sparse(self):
        self.delxattr(self._sparse_key)

    def _mark_sparse(self):
        self.setxattr(self._sparse_key, b'1')


EatPath._bind_flavours()


class LocalPath(EatPath, AugmentedPath):
    # local data about remote objects

    chunksize = 4096  # make the data generator chunksize visible externally
    _cache_class = None  # must be defined by child classes
    sysid = None  # set below

    _bind_sysid = classmethod(_bind_sysid_)

    @classmethod
    def setup(cls, cache_class, remote_class_factory):
        """ call this once to bind everything together """

        cn = self.__class__.__name__
        warnings.warn(f'{cn}.setup is deprecated please switch to RemotePath._new',
                      DeprecationWarning,
                      stacklevel=2)

        cache_class.setup(cls, remote_class_factory)

    #def __truediv__(self, other):
        # TODO need a way to get the remote relative to the anchor for this
        #if isinstance(other, aug.RemotePath):
            #super().__truediv__(other.as_path())
        #else:
            #super().__truediv__(other)

    @property
    def remote(self):
        return self.cache.remote

    @property
    def cache(self):
        # local can't make a cache because id doesn't know the remote id
        # but if there is an existing cache (duh) the it can try to get it
        # otherwise it will error (correctly)
        if not hasattr(self, '_cache'):
            try:
                self._cache_class(self)  # we don't have to assign here because cache does it
            except exc.NoCachedMetadataError as e:
                #log.error(e)
                return None
            except TypeError as e:
                if self._cache_class is None:
                    return None

        return self._cache

    @property
    def cache_id(self):
        """ abstraction violating but fast way to get cache.id """
        raise NotImplementedError('implement in subclass')

    def cache_init(self, id_or_meta, anchor=False):
        """ wow it took way too long to realize this was the way to do it >_<
            **kwargs are passed to _cache_class and _remote_class.init """

        if self.cache and self.cache.meta:
            raise exc.CacheExistsError(f'{self.cache}\n'
                                       f'{self.cache.meta}')

        elif not self.exists():
            raise ValueError(f'Cannot init a cache on a non-existent path!\n{self}')
        #elif not self.is_dir():
            #raise ValueError(f'Can only init a cache on a directory!\n{self}')

        if not isinstance(id_or_meta, PathMeta):
            id_or_meta = PathMeta(id=id_or_meta)

        cache = self._cache_class(self, meta=id_or_meta)
        if anchor:
            cache.anchorClassHere()

        return cache

    def mkdir_cache(self, remote):  # XXX hack around my idiocy / desire to not hit the network
        """ wow side effects everywhere
            given a remote, create the local folder structure
            and attach the remote metadata as cache

            I think this is implemented on LocalPath because only the
            local parent exists and the cache can only come into being after
            a folder is created since there is no point in making a symlink
            just to replace it with a folder, that just thrashes disk """

        cc = self._cache_class
        rc = cc._remote_class
        for parent in reversed(tuple(remote.parents)):
            # remote as_path is always a PurePosixPath relative to the
            # anchor and does not include the anchor
            local_path = cc.anchor.local / parent.as_path()
            if not local_path.exists():
                local_path.mkdir()
                rc(parent, cache=cc(local_path, remote=parent, meta=parent.meta))

    def mkdir_remote(self, parents=False):  # XXX hack around my idiocy / desire to not hit the network
        # FIXME this should really make a cache with a temporary id ...
        # that way we have a staging area ... but that is a bit too much for right now
        if self.cache and self.remote and self.remote.exists():
            # it is ok to check all of these to make sure that
            # things don't go stale, under most circumstances
            # self.cache will be None when this is called
            raise exc.RemotePathExistsError(self.remote)

        if self.parent.cache:
            remote = self.parent.remote._mkdir_child(self.name)
            # FIXME see the note about my dumbness wrt paths
            # the whole remote appraoch nees a complete rework so that the remotes
            # just act like paths rathern than forcing them to exist which causes all
            # sorts of awkwardness, including the RemoteMaybeExists issue
            self.mkdir(exist_ok=True)
            if remote.cache is None:
                self.cache_init(remote.meta)
                remote._cache = self.cache

            return remote

        elif parents:
            for parent in self.parents:
                c = parent.cache
                if c:
                    parts = self.relative_path_from(parent).parts
                    for p in parts:
                        print(self, parent, parts, p)
                        remote = parent.cache.remote._mkdir_child(p)
                        local = parent / p
                        local.mkdir(exist_ok=True)
                        local.cache_init(remote.meta)
                        parent = local
                        #parent = remote.cache.local  # fails due to cache is None
                        # FIXME will this work as expected ?! and just put the metadata on the
                        # existing folder or will it barf? the answer is no, cache is missing

                    remote._cache = local.cache
                    return remote
        else:
            raise FileNotFoundError('missing parent for {self} and parents=False')

    def find_cache_root(self, fail=False):
        """ find the local root of the cache tree, even if we start with skips """
        if self.is_broken_symlink():
            return self.parent.find_cache_root(fail=fail)

        found_cache = None
        # try all the variants in case some symlinking weirdness is going on
        # TODO may want to detect and warn on that?
        root = self.__class__('/')
        for variant in set((self, self.absolute(), self.resolve())):
            for parent in chain((variant,), variant.parents):
                try:
                    if parent.cache:
                        found_cache = parent
                except (exc.NoCachedMetadataError, exc.NotInProjectError) as e:
                    if fail:
                        raise e

                    # if we had a cache, went to the parent and lost it
                    # then we are at the root, assuming of course that
                    # there aren't sparse caches on the way up (down?) the tree
                    if found_cache is not None and found_cache != root:
                        return found_cache

            else:
                if found_cache and found_cache != root:
                    return found_cache

    @property
    def skip_cache(self):
        """ returns True if the path should not have a cache
            because the file has been ignored """
        if self.cache is None:
            parent_cache = self.parent.cache
            if parent_cache:
                rel_path = self.relative_to(parent_cache.anchor)
            else:
                # FIXME is it safe to cache the results of finding the root??
                root = self.find_cache_root(fail=True)
                rel_path = self.relative_to(root)
            return (rel_path.parts[0] in self._cache_class.cache_ignore or
                    # TODO more conditions
                    False)
        else:
            # FIXME technically not correct ...
            return True

    def dedupe(self, other, pretend=False):
        return self.cache.dedupe(other.cache, pretend=pretend)

    @property
    def id(self):  # FIXME reuse of the name here could be confusing, though it is technically correct
        """ THERE CAN BE ONLY ONE """
        # return self.checksum()  # doesn't quite work for folders ...
        # return self.as_posix()  # FIXME which one to use ...
        return self.sysid + ':' + self.as_posix()

    @property
    def created(self):
        self.meta.created

    def _stat(self):
        """ sometimes python just doesn't have what it takes """
        # we can't define this or use error trapping in self.stat() directly
        # because it will bolox other things that need stat to fail correctly
        cmd = ['stat', self.as_posix(), '-c', StatResult.stat_format]
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, errs = p.communicate()
        # local results retain the quotes where pexpect does not
        out = out.strip(b'"').rstrip().rstrip(b'"')
        if not out and errs:
            p.returncode
            raise OSError

        return StatResult(out)

    @property
    def size(self):
        """ don't use this to populate meta, but meta computes a checksum
            so if you need anything less than the checksum don't get meta """
        try:
            st = self.stat()
        except OSError as e:
            st = self._stat()

        return st.st_size

    @property
    def meta(self):
        return self._meta_maker()

    @property
    def meta_no_checksum(self):
        return self._meta_maker(checksum=False)

    def _meta_maker(self, *, checksum=True, chunksize=None):
        """ setting chunksize will cause an etag to be calculated """
        if not self.exists():
            return PathMeta(
                id=self.sysid + ':' + self.as_posix(),
            )

        try:
            st = self.stat()
        except OSError as e:
            st = self._stat()

        # FIXME nanos vs millis ??
        change_tuple = (fs_metadata_changed_time,
                        fs_data_modified_time) = (st.st_ctime,
                                                  st.st_mtime)

        if hasattr(self, '_meta') and self._meta is not None:
            if self.__change_tuple == change_tuple:
                return self._meta

            old_meta = self._meta  # TODO log changes?


        self.__change_tuple = change_tuple  # TODO log or no?

        updated = datetime.fromtimestamp(fs_data_modified_time, tz=timezone.utc)
        # sanity check
        # td = (datetime.fromtimestamp(fs_data_modified_time, tz=timezone.utc)
              # - datetime(1970, 1, 1, tzinfo=timezone.utc))
        # assert int(fs_data_modified_time) == int(td.total_seconds())

        # these use our internal representation of timestamps
        # the choice of how to store them in xattrs, sqlite, json, etc is handled at those interfaces
        # replace with comma since it is conformant to the standard _and_
        # because it simplifies PathMeta as_path
        mode = oct(st.st_mode)
        self._meta = PathMeta(size=st.st_size,
                              created=None,
                              updated=updated,
                              checksum=self.checksum() if checksum else None,
                              etag=self.etag() if chunksize else None,
                              chunksize=chunksize,
                              id=self.id,
                              file_id=st.st_ino,  # pretend inode number is file_id ... oh wait ...
                              user_id=st.st_uid,
                              # keep in mind that a @meta.setter
                              # will require a coverter for non-unix uids :/
                              # man use auth is all bad :/
                              gid=st.st_gid,
                              mode=mode)

        return self._meta

    @meta.setter
    def meta(self, value):
        raise TypeError('Cannot set meta on LocalPath, it is a source of metadata.')

    def _data(self, ranges=tuple(), chunksize=None):
        """ request arbitrary subsets of data from an object """

        if chunksize is None:
            chunksize = self.chunksize

        with open(self, 'rb') as f:
            if not ranges:
                ranges = (0, None),
            else:
                # TODO validate ranges
                pass

            for start, end in ranges:
                f.seek(start, 2 if start < 0 else 0)
                if end is not None:
                    total = end - start
                    if total < chunksize:
                        nchunks = 0
                        last_chunksize = total
                    elif total > chunksize:
                        nchunks, last_chunksize = divmod(total, chunksize)

                    for _ in range(nchunks):  # FIXME boundscheck ...
                        yield f.read(chunksize)

                    yield f.read(last_chunksize)

                else:
                    while True:
                        data = f.read(chunksize)  # TODO hinting
                        if not data:
                            break

                        yield data

    @property
    def data(self):
        with open(self, 'rb') as f:
            while True:
                data = f.read(self.chunksize)  # TODO hinting
                if not data:
                    break

                yield data

    @data.setter
    def data(self, generator):
        cache = self.cache
        if cache is not None:
            cmeta = cache.meta
        else:
            assert self.cache is None

        # FIXME do we touch a file, write the meta
        # and then write the data?
        # do we touch a temporary file, write the meta
        # unlink the symlink, and move the temp file in, and then write the data?
        # the order that we do this in is very important for robustness to failure
        # especially when updating a file ...
        # storing history in the symlink cache also an option?
        log.debug(f'writing to {self}')
        chunk1 = next(generator)  # if an error occurs don't open the file
        with open(self, 'wb') as f:
            f.write(chunk1)
            for chunk in generator:
                #log.debug(chunk)
                f.write(chunk)

        if cache is not None:  # FIXME cache
            if not cache.meta:
                cache.meta = cmeta  # glories of persisting xattrs :/
            # yep sometimes the xattrs get  blasted >_<
            assert cache.meta
            assert self.cache.meta

    def _data_setter(self, generator):
        """ a data setter that can be used in a chain of generators """
        log.debug(f'writing to {self}')
        chunk1 = next(generator)  # if an error occurs don't open the file
        with open(self, 'wb') as f:
            f.write(chunk1)
            yield chunk1
            for chunk in generator:
                #log.debug(chunk)
                f.write(chunk)
                yield chunk

    def copy_to(self, target, force=False, copy_cache_meta=False):
        """ copy from a the current path object to a target path """
        if type(target) != type(self):
            target = self.__class__(target)

        if not target.exists() and not target.is_symlink() or force:
            target.data = self.data
        else:
            raise exc.PathExistsError(f'{target}')

        if copy_cache_meta:
            log.debug(f'copying cache meta {self.cache.meta}')
            target.cache_init(self.cache.meta)

    def copy_from(self, source, force=False, copy_cache_meta=False):
        """ copy from a source path to the current path object """
        if type(source) != type(self):
            source = self.__class__(source)

        source.copy_to(self, force=force, copy_cache_meta=copy_cache_meta)

    @property
    def children(self):
        if self.is_dir():
            if (self.cache is not None and
                # relative paths inside may have a cache but no anchor
                # the anchor itself can be relative and have an anchor
                # so we test to see if there is an anchor first
                self.cache.anchor and
                self == self.cache.anchor.local):
                cache_ignore = self._cache_class.cache_ignore
                # implemented this way we can still use Path to navigate
                # once we are inside local data dir, though all files there
                # are skip_cache -> True
                for path in self.iterdir():
                    if path.stem in cache_ignore:
                        continue

                    yield path
            else:
                yield from self.iterdir()

    @property
    def rchildren(self):
        if (self.is_dir() and self.cache is not None and
            # relative paths inside may have a cache but no anchor
            # the anchor itself can be relative and have an anchor
            # so we test to see if there is an anchor first
            self.cache.anchor and
            self == self.cache.anchor.local):
            for path in self.children:
                yield path
                yield from path.rchildren

        else:
            yield from self.rglob('*')

    def content_different(self):
        cmeta = self.cache.meta
        if cmeta.checksum:
            return self.meta.content_different(cmeta)
        else:
            # TODO use the index for this
            # but for now just pull down the remote file
            # NOTE: this is all handled behind the scenes
            # by cache.checksum now
            return self.checksum() != self.cache.checksum()

    def diff(self):
        """ This is a bit tricky because it means that we need to
            keep a shadow copy/cache of all the downloaded files in
            operations by default... """
        raise NotImplementedError

    def meta_to_remote(self):
        # pretty sure that we don't wan't this independent of data_to_remote
        # sort of cp vs cp -a and commit date vs author date
        raise NotImplementedError
        meta = self.meta
        # FIXME how do we invalidate cache?
        self.remote.meta = meta  # this can super duper fail

    def data_to_remote(self):
        raise NotImplementedError
        self.remote.data = self.data

    def annotations_to_remote(self):
        raise NotImplementedError
        self.remote.data = self.data

    def to_remote(self):  # push could work ...
        # FIXME in theory we could have an endpoint for each of these
        # The remote will handle that?
        # this can definitely fail
        raise NotImplementedError('need the think about how to do this without causing disasters')
        self.remote.meta = self.meta
        self.remote.data = self.data
        self.remote.annotations = self.annotations

    def etag(self, chunksize):
        """ chunksize is the etag cypher chunksize which is
            different than the data generator chunksize
            etag chunksize has be implemented so that it
            works correctly with any data generator chunksize """

        if self.is_file():
            m = etag(chunksize)
            for chunk in self.data:
                m.update(chunk)

            return m.digest()

        return


LocalPath._bind_flavours()


class XopenHelper:
    @staticmethod
    def _base_helpers(pos_helpers, win_helpers):
        pos_helpers = tuple(set(pos_helpers + (XopenPosixHelper,)))
        win_helpers = tuple(set(win_helpers + (XopenWindowsHelper,)))
        return pos_helpers, win_helpers

    @classmethod
    def _bind_flavours(cls, pos_helpers=tuple(), win_helpers=tuple()):
        super()._bind_flavours(*XopenHelper._base_helpers(pos_helpers, win_helpers))


class XopenWindowsHelper(XopenHelper):
    _command = 'start'

    def xopen(self, command=None):
        """ open file using start or `command' if provided """
        if command is None or not isinstance(command, str):
            command = self._command

        process = subprocess.Popen([command, self],
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.STDOUT)


class XopenPosixHelper(XopenHelper):
    _command = 'open' if sys.platform == 'darwin' else 'xdg-open'

    def xopen(self, command=None):
        """ open file using xdg-open or `command' if provided """
        if command is None or not isinstance(command, str):
            command = self._command

        process = subprocess.Popen([command, self.as_posix()],
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.STDOUT)

        return  # FIXME this doesn't seem to update anything beyond python??

        pid = process.pid
        proc = psutil.Process(pid)
        process_window = None
        while not process_window:  # FIXME ick
            sprocs = [proc] + [p for p in proc.children(recursive=True)]
            if len(sprocs) < 2:  # xdg-open needs to call at least one more thing
                sleep(.01)  # spin a bit more slowly
                continue

            wpids = [s.pid for s in sprocs][::-1]  # start deepest work up
            # FIXME expensive to create this every time ...
            disp = Display()
            root = disp.screen().root
            children = root.query_tree().children
            #names = [c.get_wm_name() for c in children if hasattr(c, 'get_wm_name')]
            try:
                by_pid = {c.get_full_property(disp.intern_atom('_NET_WM_PID'), 0):c for c in children}
            except Xlib.error.BadWindow:
                sleep(.01)  # spin a bit more slowly
                continue

            process_to_window = {p.value[0]:c for p, c in by_pid.items() if p}
            for wp in wpids:
                if wp in process_to_window:
                    process_window = process_to_window[wp]
                    break

            if process_window:
                name = process_window.get_wm_name()
                new_name = name + ' ' + self.resolve().as_posix()[-30:]
                break  # TODO search by pid is broken, but if you can find it it will work ...
                # https://github.com/jordansissel/xdotool/issues/14 some crazy bugs there
                command = ['xdotool', 'search','--pid', str(wp), 'set_window', '--name', f'"{new_name}"']
                subprocess.Popen(command,
                                 stdout=subprocess.DEVNULL,
                                 stderr=subprocess.STDOUT)
                print(' '.join(command))
                break
                process_window.set_wm_name(new_name)
                break
            else:
                sleep(.01)  # spin a bit more slowly


class XopenPath(XopenHelper, AugmentedPath):
    pass


XopenPath._bind_flavours()

# any additional values
LocalPath._bind_sysid()
